from django.urls import include, path, re_path
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework import permissions

urlpatterns = [
    # Make this distinct from typical production values, to ensure it works dynamically
    path('', include('rgd.urls')),
]

schema_view = get_schema_view(
    openapi.Info(
        title='Resonant GeoData API',
        default_version='v1',
        description='Resonant GeoData',
        # terms_of_service='https://www.google.com/policies/terms/',
        contact=openapi.Contact(email='kitware@kitare.com'),
        license=openapi.License(name='Apache 2.0'),
    ),
    public=False,
    permission_classes=(permissions.AllowAny,),
    patterns=urlpatterns,
)

urlpatterns += [
    re_path(
        r'^swagger(?P<format>\.json|\.yaml)$',
        schema_view.without_ui(cache_timeout=0),
        name='schema-json',
    ),
    path('swagger/', schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
    path('redoc/', schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
]



from pathlib import Path

import magic
import pytest
from rgd import utility
from rgd.datastore import datastore

SKIP_FUSE = True
try:
    import simple_httpfs  # noqa

    assert Path('/tmp/rgd/http').exists()
    SKIP_FUSE = False
except (ImportError, EnvironmentError, AssertionError):
    pass

FILENAME = 'stars.png'


@pytest.fixture
def simple_url():
    return datastore.get_url(FILENAME)


@pytest.mark.skipif(SKIP_FUSE, reason='FUSE is not installed/configured.')
@pytest.mark.django_db(transaction=True)
def test_fuse_mount(simple_url):
    assert utility.precheck_fuse(simple_url)
    path = utility.url_file_to_fuse_path(simple_url)
    assert path.exists()
    mime = magic.Magic(mime=True)
    assert mime.from_file(path) == 'image/png'



from __future__ import annotations

import contextlib
from contextlib import contextmanager
from functools import wraps
import hashlib
import io
import logging
import os
from pathlib import Path
import shutil
import tempfile
from typing import TYPE_CHECKING, Any, Optional
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen
from uuid import uuid4

import boto3
from botocore import UNSIGNED
from botocore.config import Config
from django.conf import settings
from django.contrib.gis.db.models import Model
from django.core.files import File
from django.db.models.fields.files import FieldFile
from django.http import HttpResponseRedirect
from django.utils.safestring import mark_safe
from filelock import FileLock, Timeout
import psutil
from rest_framework.response import Response

if TYPE_CHECKING:
    from rgd.models import ChecksumFile

try:
    from minio_storage.storage import MinioStorage
except ImportError:
    MinioStorage = None

logger = logging.getLogger(__name__)


def get_temp_dir():
    path = Path(getattr(settings, 'RGD_TEMP_DIR', os.path.join(tempfile.gettempdir(), 'rgd')))
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_cache_dir():
    path = Path(get_temp_dir(), 'file_cache')
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_lock_dir():
    path = Path(get_temp_dir(), 'file_locks')
    path.mkdir(parents=True, exist_ok=True)
    return path


def get_file_lock(path: Path):
    """Create a file lock under the lock directory."""
    # Computes the hash using Pathlib's hash implementation on absolute path
    sha = hash(path.absolute())
    lock_path = Path(get_lock_dir(), f'{sha}.lock')
    lock = FileLock(lock_path)
    return lock


@contextmanager
def safe_urlopen(url: str, *args, **kwargs):
    with contextlib.closing(urlopen(url, *args, **kwargs)) as remote:
        yield remote


def compute_hash(handle: io.BufferedIOBase, chunk_num_blocks: int = 128):
    sha = hashlib.sha512()
    while chunk := handle.read(chunk_num_blocks * sha.block_size):
        sha.update(chunk)
    return sha.hexdigest()


def compute_checksum_file_field(field_file: FieldFile, chunk_num_blocks: int = 128):
    with field_file.open() as f:
        return compute_hash(f, chunk_num_blocks)


def compute_checksum_url(url: str, chunk_num_blocks: int = 128):
    with safe_urlopen(url) as remote:
        return compute_hash(remote, chunk_num_blocks)


def _link_url(obj: Model, field: str):
    if not getattr(obj, field, None):
        return 'No attachment'
    attr = getattr(obj, field)
    if callable(attr):
        url = attr()
    else:
        url = attr.url
    return mark_safe(f'<a href="{url}" download>Download</a>')


def get_or_create_no_commit(model: Model, defaults: Optional[dict] = None, **kwargs):
    try:
        return model.objects.get(**kwargs), False
    except model.DoesNotExist:
        if not defaults:
            defaults = {}
        defaults.update(kwargs)
        return model(**defaults), True


def get_s3_client():
    if boto3.session.Session().get_credentials():
        s3 = boto3.client('s3')
    else:
        # No credentials present (often the case in dev), use unsigned requests
        s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
    return s3


def download_url_file_to_local_path(
    url: str,
    path: str,
    num_blocks: int = 128,
    block_size: int = 128,
) -> Path:
    dest_path = Path(path)
    dest_path.parent.mkdir(parents=True, exist_ok=True)

    parsed = urlparse(url)
    if parsed.scheme == 's3':
        s3 = get_s3_client()
        with open(dest_path, 'wb') as dest_stream:
            s3.download_fileobj(
                parsed.netloc, parsed.path.lstrip('/'), dest_stream, {'RequestPayer': 'requester'}
            )
    elif parsed.scheme == 'file':
        # File available on localfilesystem
        true_path = Path(url.replace('file://', '', 1)).absolute()
        if dest_path.exists() and not os.path.islink(str(dest_path)):
            os.remove(dest_path)  # Remove file in case touched
        if not dest_path.exists():
            os.symlink(true_path, dest_path)
        # else exists and is a symlink - ASSUME it is correct
    else:
        with safe_urlopen(url) as remote, open(dest_path, 'wb') as dest_stream:
            while chunk := remote.read(num_blocks * block_size):
                dest_stream.write(chunk)
                dest_stream.flush()

    return Path(dest_path)


def precheck_fuse(url: str) -> bool:
    try:
        import simple_httpfs  # noqa
    except (ImportError, EnvironmentError):
        return False
    parsed = urlparse(url)
    if parsed.scheme not in ['https', 'http']:
        return False
    try:
        # The FUSE lib will not catch URL errors
        with safe_urlopen(url) as _:
            pass
    except HTTPError:
        return False
    return True


def url_file_to_fuse_path(url: str) -> Path:
    # Could raise ValueError within context
    # Assumes `precheck_fuse` was verified prior
    # See https://github.com/ResonantGeoData/ResonantGeoData/issues/237
    parsed = urlparse(url)
    if parsed.scheme in ['https', 'http']:
        fuse_path = url.replace('https://', '/tmp/rgd/http/') + '..'
    elif Path('/tmp/rgd/s3/').exists() and parsed.scheme == 's3':
        fuse_path = url.replace('s3://', '/tmp/rgd/s3/') + '..'
    else:
        raise ValueError(f'Scheme {parsed.scheme} not currently handled by FUSE.')
    logger.debug(f'FUSE path: {fuse_path}')
    return Path(fuse_path)


@contextmanager
def patch_internal_presign(f: FieldFile):
    """Create an environment where Minio-based `FieldFile`s construct a locally accessible presigned URL.

    Sometimes the external host differs from the internal host for Minio files (e.g. in development).
    Getting the URL in this context ensures that the presigned URL returns the correct host for the
    odd situation of accessing the file locally.

    Note
    ----
    If concerned regarding concurrent access, see https://github.com/ResonantGeoData/ResonantGeoData/issues/287

    """
    if (
        MinioStorage is not None
        and isinstance(f.storage, MinioStorage)
        and getattr(settings, 'MINIO_STORAGE_MEDIA_URL', None) is not None
    ):
        original_base_url = f.storage.base_url
        try:
            f.storage.base_url = None
            yield
        finally:
            f.storage.base_url = original_base_url
        return
    yield


@contextmanager
def output_path_helper(filename: str, output: ChecksumFile, final_name: Optional[str] = None):
    if final_name is None:
        final_name = os.path.basename(filename)
    workdir = get_temp_dir()
    with tempfile.TemporaryDirectory(dir=workdir) as tmpdir:
        output_path = os.path.join(tmpdir, filename)
        try:
            # Yield the path for the user to perform a task
            yield output_path
        except Exception as e:
            raise e
        else:
            # Save the file contents to the output field only on success
            with open(output_path, 'rb') as f:
                output.save_file_contents(f, final_name)


@contextmanager
def input_output_path_helper(
    source: ChecksumFile, output: ChecksumFile, prefix: str = '', suffix: str = ''
):
    """Yield source and output paths between a ChecksumFile and a FileFeild.

    The output path is saved to the output field after yielding.

    """
    filename = prefix + os.path.basename(source.name) + suffix
    final = Path(source.name).parent / filename
    with source.yield_local_path() as file_path:
        with output_path_helper(filename, output, final_name=final) as output_path:
            try:
                # Yield the paths for the user to perform a task
                yield (file_path, output_path)
            except Exception as e:
                raise e


def uuid_prefix_filename(instance: Any, filename: str):
    """Use a variable in settings to add a prefix to the path and keep the random uuid."""
    prefix = getattr(settings, 'RGD_FILE_FIELD_PREFIX', None)
    if prefix:
        return f'{prefix}/{uuid4()}/{filename}'
    return f'{uuid4()}/{filename}'


def skip_signal():
    """Skip the signal on an instance-basis."""

    def _skip_signal(signal_func):
        @wraps(signal_func)
        def _decorator(sender, instance, *args, **kwargs):
            if hasattr(instance, 'skip_signal') and instance.skip_signal:
                return None
            return signal_func(sender, instance, *args, **kwargs)

        return _decorator

    return _skip_signal


def download_field_file_to_local_path(field_file: FieldFile, path: str) -> Path:
    """Download entire FieldFile to disk location.

    This overrides `girder_utils.field_file_to_local_path` to download file to
    local path without a context manager. Cleanup must be handled by caller.

    """
    dest_path = Path(path)
    dest_path.parent.mkdir(parents=True, exist_ok=True)
    the_path = Path(dest_path)
    with field_file.open('rb'):
        file_obj: File = field_file.file
        if type(file_obj) is File:
            # When file_obj is an actual File, (typically backed by FileSystemStorage),
            # it is already at a stable path on disk.
            # We must symlink it into the desired path
            os.symlink(file_obj.name, dest_path)
            return the_path
        else:
            # When file_obj is actually a subclass of File, it only provides a Python
            # file-like object API. So, it must be copied to a stable path.
            with open(dest_path, 'wb') as dest_stream:
                shutil.copyfileobj(file_obj, dest_stream)
                dest_stream.flush()
            return the_path


def clean_file_cache(override_target=None):
    """Clean the file cache to achieve RGD_TARGET_AVAILABLE_CACHE (Gb).

    Note
    ----
    If above target, this will not do anything.

    Note
    ----
    This is blocking across threads/processes; only one clean can happen at a
    time.

    Return
    ------
    A tuple of the starting and ending free space in bytes.

    """
    cache = get_cache_dir()
    # While free space is not enough, remove directories until all ar gone
    initial = psutil.disk_usage(cache).free
    target = override_target or getattr(settings, 'RGD_TARGET_AVAILABLE_CACHE', 2)
    if psutil.disk_usage(cache).free * 1e-9 >= target:
        # We're above target, so immediately return
        return initial, psutil.disk_usage(cache).free
    # Below target, starting a clean - this is blocking across processes
    cache_lock = get_file_lock(cache)
    with cache_lock:
        logger.debug(f'Cleaning file cache... Starting free space is {initial} bytes.')
        # Sort each directory by mtime
        paths = sorted(Path(cache).iterdir(), key=os.path.getmtime)  # This sorts oldest to latest
        while psutil.disk_usage(cache).free * 1e-9 < target:
            if not len(paths):
                # If we delete everything and still cannot achieve target, warn
                logger.error(
                    f'Target cache free space of {target * 1e9} bytes not achieved when empty.'
                )
                break
            path = paths.pop(0)
            # Check if resource is locked and skip if so
            lock = get_file_lock(path)
            try:
                with lock.acquire(timeout=0):
                    if os.path.isdir(path):
                        shutil.rmtree(path)
                    else:
                        os.remove(path)
                # Remove the lockfile as well
                os.remove(lock.lock_file)
                logger.debug(f'removed: {path}')
            except Timeout:
                # Another task holds the lock on this file/directory: do not delete
                logger.debug(f'File is locked, skipping: {path}')
                pass
    free = psutil.disk_usage(cache).free
    logger.debug(f'Finished cleaning file cache. Available free space is {free} bytes.')
    return initial, free


def purge_file_cache():
    """Completely purge all files from the file cache.

    This should be used with extreme caution, it will delete files that could
    be in use.

    """
    cache = get_temp_dir()
    shutil.rmtree(cache)
    cache = get_cache_dir()  # Return the cache dir so that a fresh directory is created.
    logger.debug(
        f'Purged file cache. Available free space is {psutil.disk_usage(cache).free} bytes.'
    )
    return cache


def get_file_data_url(file: ChecksumFile) -> HttpResponseRedirect | Response:
    # Import here to avoid circular dependency
    from rgd.models.file import FileSourceType

    if file.type == FileSourceType.FILE_FIELD:
        return HttpResponseRedirect(file.get_url())
    elif file.type == FileSourceType.URL:
        return Response(file.url)
    else:
        raise NotImplementedError('Invalid file type')



import inspect

from django.contrib.auth.models import User
import factory
import pytest
from pytest_factoryboy import register
from rest_framework.authtoken.models import Token
from rgd_client import create_rgd_client
from rgd_testing_utils import factories
from rgd_testing_utils.api_fixtures import *  # noqa
from rgd_testing_utils.data_fixtures import *  # noqa

for _, fac in inspect.getmembers(factories):
    if inspect.isclass(fac) and issubclass(fac, factory.Factory):
        register(fac)


@pytest.fixture
def py_client(live_server):
    params = {'username': 'test@kitware.com', 'email': 'test@kitware.com', 'password': 'password'}

    user = User.objects.create_user(is_staff=True, is_superuser=True, **params)
    user.save()

    # use constant value for API key so this client fixture can be reused across multiple tests
    Token.objects.create(user=user, key='topsecretkey')

    client = create_rgd_client(
        username=params['username'], password=params['password'], api_url=f'{live_server.url}/api'
    )

    return client


@pytest.fixture
def user_with_api_key(faker):
    email = faker.email()
    password = 'password'
    params = {'username': email, 'email': email, 'password': password}

    user = User.objects.create_user(is_staff=True, is_superuser=True, **params)
    user.save()

    api_token = 'topsecretkey'
    Token.objects.create(user=user, key=api_token)

    return email, password, api_token



from django.urls import path
from rest_framework.authtoken.views import obtain_auth_token
from rest_framework.routers import SimpleRouter
from rgd import views
from rgd.rest import viewsets

router = SimpleRouter(trailing_slash=False)
router.register(r'api/rgd/collection', viewsets.CollectionViewSet)
router.register(r'api/rgd/collection_permission', viewsets.CollectionPermissionViewSet)
router.register(r'api/rgd/checksum_file', viewsets.ChecksumFileViewSet)
router.register(r'api/rgd/spatial_entry', viewsets.SpatialEntryViewSet)
router.register(r'api/rgd/spatial_asset', viewsets.SpatialAssetViewSet)

urlpatterns = [
    # API Key Authentication
    path('api/api-token-auth', obtain_auth_token, name='api-token-auth'),
    path('api/signature', viewsets.SignatureView.as_view(), name='signature'),
    # Pages
    path('', views.SpatialEntriesListView.as_view(), name='rgd-index'),
    path(
        'rgd/statistics',
        views.StatisticsView.as_view(),
        name='statistics',
    ),
    path(
        'rgd/spatial_entries/<int:pk>/',
        views.spatial_entry_redirect_view,
        name='spatial-entry-detail',
    ),
    # Deprecated
    # TODO: remove route. duplicated functionality in 'rgd.viewsets.SpatialEntryViewSet'
    path(
        'api/rgd/search',
        viewsets.SpatialEntryViewSet.as_view({'get': 'list'}),
    ),
] + router.urls



from contextlib import contextmanager
import io
import logging
from pathlib import Path
from typing import List, Union
from urllib.parse import urlparse

from django.core.exceptions import ValidationError
from django.db.models import QuerySet

from ..utility import compute_checksum_url, compute_hash, get_file_lock, get_or_create_no_commit
from .collection import Collection
from .file import ChecksumFile, FileSourceType
from .mixins import Status

logger = logging.getLogger(__name__)


def get_or_create_checksum_file_url(
    url: str,
    collection: Collection = None,
    precompute_url_checksum: bool = False,
    defaults: dict = None,
    **kwargs,
):
    """Get or create ChecksumFile from unique URL and Collection.

    Parameters
    ----------
    precompute_url_checksum : bool
        Precompute the checksum of URL files to prevent duplicate files at different addresses.

    """
    if defaults is None:
        defaults = {}
    parsed = urlparse(url)
    if parsed.scheme not in ['https', 'http', 's3']:
        raise ValidationError(f'Not a supported URL scheme ({parsed.scheme}) for URL: {url}')
    if precompute_url_checksum:
        checksum = compute_checksum_url(url)
        kwargs.setdefault('checksum', checksum)
    try:
        file_entry = ChecksumFile.objects.get(url=url, collection=collection, **kwargs)
        if file_entry.status != Status.SUCCEEDED and file_entry.status != Status.SKIPPED:
            # If in a failed state, rerun tasks
            file_entry.save()
        return file_entry, False
    except ChecksumFile.DoesNotExist:
        defaults = defaults.copy()
        defaults.update(kwargs)
        file_entry = ChecksumFile(**defaults)
        file_entry.url = url
        file_entry.type = FileSourceType.URL
        file_entry.collection = collection
        file_entry.save()
        return file_entry, True


def get_or_create_checksumfile_file_field(
    file_handle: io.BufferedIOBase, collection: Collection = None, defaults: dict = None, **kwargs
):
    """Get or create ChecksumFile from an open file handle and Collection."""
    if 'checksum' not in kwargs:
        file_handle.seek(0)
        checksum = compute_hash(file_handle)
        kwargs.setdefault('checksum', checksum)
    file_entry, created = get_or_create_no_commit(
        ChecksumFile, collection=collection, defaults=defaults, **kwargs
    )
    if created:
        # Since we already computed the checksum, prevent tasks from recomputing
        file_entry.skip_signal = True
        file_handle.seek(0)
        file_entry.save_file_contents(file_handle, 'test')
        file_entry.save()
        file_entry.skip_signal = False
    return file_entry, created


def get_or_create_checksumfile(
    collection: Collection = None, precompute_url_checksum=False, defaults: dict = None, **kwargs
):
    """Get or create a ChecksumFile.

    Parameters
    ----------
    precompute_url_checksum : bool
        Precompute the checksum of URL files to prevent duplicate files at different addresses.

    Notes
    -----
    * URL files are only evaluated by their URL and Collection. Use the `url` keyword argument.
    * FILE_FIELD files are evaluated by their checksum value and Collection. Use the `file` keyword argument.

    Since files cannot belong to multiple Collections, we have to validate
    against a Collection. If no Collection is given, then the Collection
    of any existing duplicate file must be None. Otherwise, the file will be
    created to prevent cross-collection permission issues.

    """
    if defaults is None:
        defaults = {}
    if 'collection' in kwargs or 'collection' in defaults:
        raise ValueError('Multiple values of `collection` given.')
    if 'url' in kwargs:
        url = kwargs.pop('url')
        file_entry, created = get_or_create_checksum_file_url(
            url,
            collection=collection,
            precompute_url_checksum=precompute_url_checksum,
            defaults=defaults,
            **kwargs,
        )
    elif 'file' in kwargs:
        file_handle = kwargs.pop('file')
        file_entry, created = get_or_create_checksumfile_file_field(
            file_handle, collection=collection, defaults=defaults, **kwargs
        )
    else:
        raise ValidationError('No URL or file contents passed.')
    return file_entry, created


@contextmanager
def yield_checksumfiles(queryset: Union[QuerySet, List[ChecksumFile]], directory: str):
    """Checkout a queryset of ChecksumFile records under a single directory.

    This will use the `name` field of each of the files as their relative path
    under the temporary directory.

    Please note that this uses a contextmanager to acquire a lock on the
    directory to make sure the files are not automatically cleaned up by
    other threads or processes.

    """
    files = list(queryset) if isinstance(queryset, QuerySet) else queryset
    # Touch the directory to update the mtime
    directory = Path(directory)
    directory.touch()
    # Acquire a lock on the directory so that it isn't cleaned up
    lock = get_file_lock(directory)
    lock.acquire()
    # Download each file to the directory and yield it so that the lock can be released when done
    try:
        names = set()
        # TODO: implement a FUSE interface
        for file in files:
            if file.name in names:
                # NOTE: caller's responsibility to handle duplicate names
                logger.error(
                    f'Duplicate `name` for ChecksumFile ({file.pk}: {file.name}). Overwriting...'
                )
            names.add(file.name)
            file.download_to_local_path(directory=directory)
        yield directory
    finally:
        lock.release()



from datetime import datetime

import pytest
from rgd import models
from rgd.filters import SpatialEntryFilter


@pytest.mark.django_db(transaction=True)
def test_q_distance(spatial_asset_a, spatial_asset_b):
    assert models.SpatialEntry.objects.count() == 2
    # Make sure all are returned sorted by distance
    filterset = SpatialEntryFilter(
        data={
            'q': f'SRID=4326;{spatial_asset_a.outline.wkt}',
        }
    )
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 2
    assert qs.first().spatial_id == spatial_asset_a.spatial_id
    filterset = SpatialEntryFilter(
        data={
            'q': f'SRID=4326;{spatial_asset_b.outline.wkt}',
        }
    )
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 2
    assert qs.first().spatial_id == spatial_asset_b.spatial_id


@pytest.mark.django_db(transaction=True)
def test_footprint_intersects(spatial_asset_a, spatial_asset_b):
    assert models.SpatialEntry.objects.count() == 2
    filterset = SpatialEntryFilter(
        data={
            'q': f'SRID=4326;{spatial_asset_a.outline.wkt}',
            'predicate': 'intersects',
        }
    )
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 1


@pytest.mark.django_db(transaction=True)
def test_geojson_intersects(spatial_asset_a, spatial_asset_b):
    assert models.SpatialEntry.objects.count() == 2
    filterset = SpatialEntryFilter(
        data={
            'q': f'{spatial_asset_a.outline.geojson}',
            'predicate': 'intersects',
        }
    )
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 1
    assert qs.first().spatial_id == spatial_asset_a.spatial_id


@pytest.mark.django_db(transaction=True)
def test_q_time(spatial_asset_a):
    spatial_asset_a.acquisition_date = datetime(2021, 1, 1, hour=12, second=0)
    spatial_asset_a.save()
    filterset = SpatialEntryFilter(data={'time_of_day_before': '13:00'})
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 1
    filterset = SpatialEntryFilter(data={'time_of_day_after': '13:00'})
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 0


@pytest.mark.django_db(transaction=True)
def test_collection(spatial_asset_a):
    filterset = SpatialEntryFilter(
        data={'collections': [spatial_asset_a.files.first().collection.pk]}
    )
    assert filterset.is_valid()
    qs = filterset.filter_queryset(models.SpatialEntry.objects.all())
    assert qs.count() == 1
    assert spatial_asset_a.spatial_id == qs.first().spatial_id



from __future__ import annotations

from datetime import timedelta
import logging
import os
import tempfile
from typing import Type

try:
    from composed_configuration import ComposedConfiguration, ConfigMixin
    from configurations import values
except ImportError:
    raise ImportError(
        'Please install `django-composed-configuration` and `django-configurations` '
        'to use the configuration mixins. This can be done through the `configuration` '
        'extra when installing `django-rgd`.'
    )


class GeoDjangoMixin(ConfigMixin):
    @staticmethod
    def mutate_configuration(configuration: Type[ComposedConfiguration]):
        configuration.INSTALLED_APPS += ['django.contrib.gis']

        try:
            import osgeo

            configuration.GDAL_LIBRARY_PATH = osgeo.GDAL_LIBRARY_PATH
            configuration.GEOS_LIBRARY_PATH = osgeo.GEOS_LIBRARY_PATH
        except (ImportError, AttributeError):
            logging.warning(
                'GDAL wheel not installed, skipping configuration. If you have not '
                'installed GDAL manually, please install the wheel with the following command: '
                'pip install --find-links https://girder.github.io/large_image_wheels GDAL'
            )


class SwaggerMixin(ConfigMixin):
    REFETCH_SCHEMA_WITH_AUTH = True
    REFETCH_SCHEMA_ON_LOGOUT = True
    OPERATIONS_SORTER = 'alpha'
    DEEP_LINKING = True


class RedisCacheMixin(ConfigMixin):
    SESSION_ENGINE = 'django.contrib.sessions.backends.cached_db'
    CACHES = values.CacheURLValue(
        default='locmem://',
        environ_name='REDIS_URL',
        environ_prefix='DJANGO',
    )


class ResonantGeoDataBaseMixin(GeoDjangoMixin, SwaggerMixin, ConfigMixin):
    @staticmethod
    def mutate_configuration(configuration: ComposedConfiguration) -> None:
        configuration.MIDDLEWARE += [
            'crum.CurrentRequestUserMiddleware',
        ]
        configuration.REST_FRAMEWORK['DEFAULT_AUTHENTICATION_CLASSES'] += [
            'rest_framework.authentication.TokenAuthentication',
        ]

    @classmethod
    def post_setup(cls):
        super().post_setup()
        if getattr(cls, 'DEBUG', False) or cls.RGD_DEBUG_LOGS:
            cls.LOGGING['loggers']['rgd'] = {
                'level': 'DEBUG',
                'handlers': ['console'],
                'propagate': False,
            }

    # This cannot have a default value, since the password and database name are always
    # set by the service admin
    DATABASES = values.DatabaseURLValue(
        environ_name='DATABASE_URL',
        environ_prefix='DJANGO',
        environ_required=True,
        # Additional kwargs to DatabaseURLValue are passed to dj-database-url
        engine='django.contrib.gis.db.backends.postgis',
        conn_max_age=600,
    )

    CELERY_WORKER_SEND_TASK_EVENTS = True
    CELERY_TASK_TIME_LIMIT = values.IntegerValue(
        environ=True, default=timedelta(days=1).total_seconds()
    )

    RGD_FILE_FIELD_PREFIX = values.Value(default=None)
    RGD_GLOBAL_READ_ACCESS = values.Value(default=False)
    RGD_AUTO_APPROVE_SIGN_UP = values.Value(default=False)
    RGD_AUTO_COMPUTE_CHECKSUMS = values.Value(default=False)
    RGD_TEMP_DIR = values.Value(default=os.path.join(tempfile.gettempdir(), 'rgd'))
    RGD_TARGET_AVAILABLE_CACHE = values.Value(default=2)
    RGD_REST_CACHE_TIMEOUT = values.Value(default=60 * 60 * 2)
    RGD_SIGNED_URL_TTL = values.Value(default=60 * 60 * 24)  # 24 hours
    RGD_SIGNED_URL_QUERY_PARAM = values.Value(default='signature')
    RGD_DEBUG_LOGS = values.Value(default=True)



import json

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.gis.db.models import Collect, Extent
from django.contrib.gis.db.models.functions import AsGeoJSON, Centroid
from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models import Count, Max, Min, Q
from django.shortcuts import redirect
from django.views import generic
from django.views.generic import DetailView, TemplateView
from rest_framework.reverse import reverse
from rgd import permissions

from . import filters, models


class PermissionTemplateView(LoginRequiredMixin, TemplateView):
    pass


class PermissionDetailView(LoginRequiredMixin, DetailView):
    def get_object(self):
        obj = super().get_object()
        permissions.check_read_perm(self.request.user, obj)
        return obj


class PermissionListView(LoginRequiredMixin, generic.ListView):
    def get_queryset(self):
        queryset = super().get_queryset()
        return permissions.filter_read_perm(self.request.user, queryset)


def query_params(params):
    query = params.copy()

    if query.get('page'):
        del query['page']

    return '&' + query.urlencode() if query.urlencode() else ''


class SpatialListView(PermissionListView):
    paginate_by = 15

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        full_queryset = self.object_list
        paginated_queryset = context['object_list']
        count = full_queryset.count()
        summary = paginated_queryset.aggregate(
            Collect('outline'),
            Extent('outline'),
            Count('pk'),
        )
        # Have a smaller dict of meta fields to parse for menu bar
        # This keeps us from parsing long GeoJSON fields twice
        context['extents_meta'] = json.dumps({'count': summary['pk__count']})
        extents = {'count': count}
        if count:
            extents.update(
                {
                    'collect': json.loads(summary['outline__collect'].geojson),
                    'convex_hull': json.loads(summary['outline__collect'].convex_hull.geojson),
                    'extent': {
                        'xmin': summary['outline__extent'][0],
                        'ymin': summary['outline__extent'][1],
                        'xmax': summary['outline__extent'][2],
                        'ymax': summary['outline__extent'][3],
                    },
                }
            )
        context['extents'] = json.dumps(extents)
        context['search_params'] = json.dumps(self.request.GET)
        context['query_params'] = query_params(self.request.GET)
        return context


class SpatialEntriesListView(SpatialListView):
    queryset = models.SpatialEntry.objects.select_subclasses()
    context_object_name = 'spatial_entries'
    template_name = 'rgd/spatialentry_list.html'
    filterset_class = filters.SpatialEntryFilter

    def get_queryset(self):
        queryset = super().get_queryset()
        filterset = self.filterset_class(data=self.request.GET)
        if not filterset.is_valid():
            message = 'Filter parameters illformed. Full results returned.'
            all_error_messages_content = [
                msg.message
                for msg in list(messages.get_messages(self.request))
                if msg.level_tag == 'error'
            ]
            if message not in all_error_messages_content:
                messages.add_message(self.request, messages.ERROR, message)
        else:
            queryset = filterset.filter_queryset(queryset)
        return queryset


class StatisticsView(PermissionListView):
    paginate_by = None
    model = models.SpatialEntry
    context_object_name = 'spatial_entries'
    template_name = 'rgd/statistics.html'

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        context.update(
            self.object_list.aggregate(
                count=Count('spatial_id', distinct=True),
                coordinates=ArrayAgg(AsGeoJSON(Centroid('footprint'))),
                instrumentation_count=Count(
                    'instrumentation',
                    distinct=True,
                    filter=Q(instrumentation__isnull=False),
                ),
                acquisition_date__min=Min(
                    'acquisition_date', filter=Q(acquisition_date__isnull=False)
                ),
                acquisition_date__max=Max(
                    'acquisition_date', filter=Q(acquisition_date__isnull=False)
                ),
                extents=Extent('outline'),
            )
        )
        context['coordinates'] = '[' + ','.join(context['coordinates']) + ']'
        extents = context['extents'] or [None] * 4
        context['extents'] = json.dumps(
            {
                'xmin': extents[0],
                'ymin': extents[1],
                'xmax': extents[2],
                'ymax': extents[3],
            }
        )
        return context


class SpatialDetailView(PermissionDetailView):
    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        extents = {'count': 0}
        if self.object.footprint:
            extents.update(
                {
                    'count': 1,
                    'collect': self.object.footprint.json,
                    'outline': self.object.outline.json,
                    'extent': self.object.bounds,
                }
            )
        context['extents'] = json.dumps(extents)
        return context


def spatial_entry_redirect_view(request, pk):
    # NOTE HACK: this is dirty but it works
    spat = models.SpatialEntry.objects.filter(pk=pk).select_subclasses().first()
    if hasattr(spat, 'detail_view_pk'):
        pk = spat
        for f in spat.detail_view_pk.split('__'):
            pk = getattr(pk, f)
    else:
        pk = spat.pk
    return redirect(reverse(spat.detail_view_name, kwargs={'pk': pk}))



from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.models import FileSet

from .mixins import MODIFIABLE_FILTERS


@admin.register(FileSet)
class FileSetAdmin(OSMGeoAdmin):
    list_display = [
        'pk',
        'name',
    ]
    list_filter = MODIFIABLE_FILTERS



"""Django REST framework style mixins for views and viewsets."""
from django_filters.rest_framework import DjangoFilterBackend
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import action
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rgd.permissions import CollectionAuthorization, CollectionAuthorizationFilter


class BaseRestViewMixin:
    permission_classes = [CollectionAuthorization, IsAuthenticated]
    filter_backends = [CollectionAuthorizationFilter, DjangoFilterBackend]


class TaskEventViewSetMixin:
    """
    Mixin for any TaskEventMixin model.

    Provides the `status` action to retrieve the status of a
    model's task event.

    TODO: This should use a serializer instead.
    """

    @swagger_auto_schema(
        method='GET',
        operation_summary='Check the status.',
    )
    @action(detail=True)
    def status(self, *args, **kwargs):
        obj = self.get_object()
        data = {
            'pk': obj.pk,
            'model': type(obj).__name__,
            'status': obj.status,
        }
        return Response(data)



import traceback

from celery.utils.log import get_task_logger

logger = get_task_logger(__name__)


def _safe_execution(func, *args, **kwargs):
    """Execute a task and return any tracebacks that occur as a string."""
    try:
        func(*args, **kwargs)
        return ''
    except Exception as exc:
        logger.exception(f'Internal error run `{func.__name__}`: {exc}')
        return traceback.format_exc()


def _run_with_failure_reason(model, func, *args, **kwargs):
    """Run a function that will update the model's `failure_reason`."""
    from rgd.models.mixins import Status

    model.status = Status.RUNNING
    model.save(update_fields=['status'])
    model.failure_reason = _safe_execution(func, *args, **kwargs)
    if model.failure_reason:
        model.status = Status.FAILED
    else:
        model.status = Status.SUCCEEDED
    model.save(update_fields=['failure_reason', 'status'])



from celery import shared_task
from django.conf import settings

from . import helpers


@shared_task(time_limit=settings.CELERY_TASK_TIME_LIMIT)
def task_checksum_file_post_save(checksumfile_pk):
    from rgd.models import ChecksumFile
    from rgd.models.mixins import Status

    obj = ChecksumFile.objects.get(pk=checksumfile_pk)

    if obj.validate_checksum or getattr(settings, 'RGD_AUTO_COMPUTE_CHECKSUMS', False):
        helpers._run_with_failure_reason(obj, obj.post_save_job)
    else:
        obj.status = Status.SKIPPED
        obj.save(update_fields=['status'])



import os
import tempfile

import psutil
import pytest
from rgd import utility
from rgd.datastore import datastore
from rgd.models import ChecksumFile, FileSet, FileSourceType, utils

FILENAME = 'stars.png'


@pytest.fixture
def file_path():
    return datastore.fetch(FILENAME)


@pytest.mark.parametrize(
    'filename',
    (FILENAME, f'a/b/c/{FILENAME}'),  # edge case - file name that represents a file path
)
@pytest.mark.django_db(transaction=True)
def test_yield_local_path_file(file_path, filename):
    model = ChecksumFile()
    model.type = FileSourceType.FILE_FIELD
    with open(file_path, 'rb') as f:
        model.file.save(filename, f)
        model.name = filename
    model.file_set = FileSet.objects.create()
    model.save()
    path = model.yield_local_path()
    with model.yield_local_path() as path:
        assert os.path.exists(path)


@pytest.mark.parametrize(
    'filename',
    (FILENAME, f'a/b/c/{FILENAME}'),  # edge case - file name that represents a file path
)
@pytest.mark.django_db(transaction=True)
def test_yield_local_path_url_http(filename):
    model = ChecksumFile()
    model.type = FileSourceType.URL
    model.url = datastore.get_url(FILENAME)
    model.name = filename
    model.save()
    with model.yield_local_path() as path:
        assert os.path.exists(path)


@pytest.mark.parametrize(
    'filename',
    (FILENAME, f'a/b/c/{FILENAME}'),  # edge case - file name that represents a file path
)
@pytest.mark.django_db(transaction=True)
def test_yield_local_path_url_s3(s3_url, filename):
    model = ChecksumFile()
    model.type = FileSourceType.URL
    model.url = s3_url
    model.name = filename
    model.save()
    with model.yield_local_path() as path:
        assert os.path.exists(path)


@pytest.mark.django_db(transaction=True)
def test_yield_checksumfiles(s3_url):
    # This tests downloading ChecksumFiles not in a FileSet to the same directory
    # Two URL files
    url = datastore.get_url('afie_1.jpg')
    file_1, _ = utils.get_or_create_checksumfile(url=url, name='afie_1.jpg')
    url = datastore.get_url('afie_2.jpg')
    file_2, _ = utils.get_or_create_checksumfile(url=url, name='the/best/dog/afie_2.jpeg')
    # One S3 URL file
    file_3, _ = utils.get_or_create_checksumfile(url=s3_url, name='s3/file/stuff.json')
    # One FileField file
    with open(datastore.fetch('afie_3.jpg'), 'rb') as f:
        file_4, _ = utils.get_or_create_checksumfile(file=f, name='afie_3.jpg')
    # Checkout all of these files under a single temporary directory
    # Note that 2 files are at top level and one file is nested
    files = ChecksumFile.objects.all()
    assert files.count() == 4
    with tempfile.TemporaryDirectory() as tempdir:
        with utils.yield_checksumfiles(files, tempdir) as directory:
            assert os.path.exists(directory)
            for f in files.all():
                assert os.path.exists(os.path.join(directory, f.name))


@pytest.mark.django_db(transaction=True)
def test_yield_checksumfiles_file_set(s3_url):
    file_set = FileSet.objects.create()
    # Two URL files
    url = datastore.get_url('afie_1.jpg')
    file_1, _ = utils.get_or_create_checksumfile(url=url, name='afie_1.jpg', file_set=file_set)
    url = datastore.get_url('afie_2.jpg')
    file_2, _ = utils.get_or_create_checksumfile(
        url=url, name='the/best/dog/afie_2.jpeg', file_set=file_set
    )
    # One S3 URL file
    file_3, _ = utils.get_or_create_checksumfile(
        url=s3_url, name='s3/file/stuff.json', file_set=file_set
    )
    # One FileField file
    with open(datastore.fetch('afie_3.jpg'), 'rb') as f:
        file_4, _ = utils.get_or_create_checksumfile(file=f, name='afie_3.jpg', file_set=file_set)

    # Checkout all of these files under a single temporary directory through the file_set
    # Note that 2 files are at top level and one file is nested
    files = ChecksumFile.objects.all()
    assert files.count() == 4
    with file_set.yield_all_to_local_path() as directory:
        assert os.path.exists(directory)
        for f in files.all():
            assert os.path.exists(os.path.join(directory, f.name))


@pytest.mark.django_db(transaction=True)
def test_clean_file_cache(checksum_file, checksum_file_url):
    f = FileSet.objects.create()
    checksum_file.file_set = f
    checksum_file.save(
        update_fields=[
            'file_set',
        ]
    )
    f.refresh_from_db()

    # Make sure clean_file_cache does not clean files in use
    with checksum_file_url.yield_local_path(yield_file_set=True) as path:
        assert os.path.exists(path)  # Make sure the file was checked out
        # Set target to size of drive to delete entire cache
        utility.clean_file_cache(override_target=psutil.disk_usage('/').total)
        assert os.path.exists(path)  # Make sure file is still present

    # Make sure the same works for `FileSet`s
    with f.yield_all_to_local_path() as directory:
        assert os.path.exists(directory)
        assert os.path.exists(checksum_file.get_cache_path())
        utility.clean_file_cache(override_target=psutil.disk_usage('/').total)
        assert os.path.exists(directory)
        assert os.path.exists(checksum_file.get_cache_path())
    # Checkout the file_set through a contained file
    with checksum_file.yield_local_path(yield_file_set=True) as path:
        assert os.path.exists(path)
        utility.clean_file_cache(override_target=psutil.disk_usage('/').total)
        assert os.path.exists(path)



from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.models import ChecksumFile

from .mixins import MODIFIABLE_FILTERS, TASK_EVENT_FILTERS, TASK_EVENT_READONLY, reprocess


def validate_checksum(modeladmin, request, queryset):
    for file in queryset.all():
        file.validate_checksum = True
        file.save()


@admin.register(ChecksumFile)
class ChecksumFileAdmin(OSMGeoAdmin):
    list_display = (
        'pk',
        'name',
        'type',
        'status',
        'created',
        'created_by',
        'modified',
        'collection',
        'data_link',
    )
    readonly_fields = (
        'checksum',
        'last_validation',
    ) + TASK_EVENT_READONLY
    actions = (
        reprocess,
        validate_checksum,
    )
    list_filter = (
        MODIFIABLE_FILTERS
        + TASK_EVENT_FILTERS
        + (
            'type',
            'collection',
            'created_by',
        )
    )
    raw_id_fields = (
        'file_set',
        'collection',
        'created_by',
    )



# Generated by Django 3.2.8 on 2021-11-05 17:15

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0004_alter_checksumfile_status'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileSet',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                (
                    'created',
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name='created'
                    ),
                ),
                (
                    'modified',
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name='modified'
                    ),
                ),
                ('name', models.CharField(blank=True, max_length=1000, null=True)),
                ('description', models.TextField(blank=True, null=True)),
            ],
            options={
                'get_latest_by': 'modified',
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='checksumfile',
            name='file_set',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to='rgd.fileset',
            ),
        ),
        migrations.AddConstraint(
            model_name='checksumfile',
            constraint=models.UniqueConstraint(fields=('file_set', 'name'), name='unique_name'),
        ),
    ]



from django.conf import settings
from django.shortcuts import get_object_or_404
from drf_yasg.utils import swagger_auto_schema
from rest_framework import response, views
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rgd import models, serializers
from rgd.filters import CollectionFilter, SpatialEntryFilter
from rgd.models.file import ChecksumFile
from rgd.rest.base import ModelViewSet, ReadOnlyModelViewSet
from rgd.utility import get_file_data_url

from .authentication import UserSigner
from .mixins import BaseRestViewMixin, TaskEventViewSetMixin


class CollectionViewSet(ModelViewSet):
    serializer_class = serializers.CollectionSerializer
    queryset = models.Collection.objects.all()
    filterset_class = CollectionFilter

    @swagger_auto_schema(
        method='GET',
        operation_summary='Get the file at the given index in this collection. Associated files are ordered by primary key and the provided index is the index in that ordered set. If the files in the collection change, this will not be reproducible.',
    )
    @action(
        detail=True,
        serializer_class=serializers.ChecksumFileSerializer,
        url_path=r'item/(?P<index>\d+)',
    )
    def item(self, request, pk, index):
        collection = get_object_or_404(models.Collection, pk=pk)
        files = collection.checksumfiles.order_by('pk')
        try:
            instance = files[int(index)]
        except (IndexError, ValueError):
            raise ValidationError(f'index {index} not valid or out of range.')
        return Response(serializers.ChecksumFileSerializer(instance).data)


class CollectionPermissionViewSet(ModelViewSet):
    serializer_class = serializers.CollectionPermissionSerializer
    queryset = models.CollectionPermission.objects.all()


class ChecksumFileViewSet(ModelViewSet, TaskEventViewSetMixin):
    serializer_class = serializers.ChecksumFileSerializer
    queryset = models.ChecksumFile.objects.all()

    @swagger_auto_schema(query_serializer=serializers.ChecksumFileListQuerySerializer())
    def list(self, request, *args, **kwargs):
        query_serializer = serializers.ChecksumFileListQuerySerializer(
            data=self.request.query_params
        )
        query_serializer.is_valid(raise_exception=True)
        url: str = query_serializer.validated_data.get('url')
        collection: int = query_serializer.validated_data.get('collection')

        if url is None and collection is None:
            return super().list(request, *args, **kwargs)

        queryset = self.get_queryset()

        if url is not None:
            queryset = queryset.filter(url=url)
        if collection is not None:
            queryset = queryset.filter(collection=collection)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    @swagger_auto_schema(
        method='GET',
        operation_summary='Download ChecksumFile data directly from S3.',
    )
    @action(detail=True)
    def data(self, request, pk=None):
        obj: ChecksumFile = self.get_object()
        return get_file_data_url(obj)

    @swagger_auto_schema(
        query_serializer=serializers.ChecksumFilePathQuerySerializer(),
        responses={200: serializers.ChecksumFilePathsSerializer()},
    )
    @action(detail=False, methods=['GET'])
    def tree(self, request, **kwargs):
        """View ChecksumFiles in a hierarchy, specifying folder/file name with path_prefix."""
        path_prefix: str = self.request.query_params.get('path_prefix') or ''
        qs = self.get_queryset().filter(name__startswith=path_prefix)

        folders: dict[str, dict] = {}
        files: dict[str, models.ChecksumFile] = {}

        for file in qs:
            file: models.ChecksumFile

            # Get the remainder of the path after path_prefix
            base_path: str = file.name[len(path_prefix) :].strip('/')

            # Since we stripped slashes, any remaining slashes indicate a folder
            folder_index = base_path.find('/')
            is_folder = folder_index >= 0

            if not is_folder:
                # Ensure base_path is entire filename
                base_path = file.name[file.name.rfind('/') + 1 :]
                files[base_path] = file
            else:
                base_path = base_path[:folder_index]
                entry = folders.get(base_path)
                fixed_file_size = file.size or 0
                url_file_as_int = 1 - int(bool(file.size))

                # Either create new folder entry, or add to existing folder
                if entry is None:
                    # New folder entry
                    folders[base_path] = {
                        'known_size': fixed_file_size,
                        'num_files': 1,
                        'num_url_files': url_file_as_int,
                        'created': file.created,
                        'modified': file.modified,
                    }
                else:
                    # Add to existing folder
                    entry['known_size'] += fixed_file_size
                    entry['num_files'] += 1
                    entry['num_url_files'] += url_file_as_int
                    entry['created'] = min(entry['created'], file.created)  # earliest
                    entry['modified'] = max(entry['modified'], file.modified)  # latest

        return Response(
            serializers.ChecksumFilePathsSerializer({'folders': folders, 'files': files}).data
        )


class SpatialEntryViewSet(ReadOnlyModelViewSet):
    serializer_class = serializers.SpatialEntrySerializer
    queryset = models.SpatialEntry.objects.select_subclasses()
    filterset_class = SpatialEntryFilter

    @swagger_auto_schema(
        method='GET',
        operation_summary='Get the footprint of this SpatialEntry.',
    )
    @action(detail=True, serializer_class=serializers.SpatialEntryFootprintSerializer)
    def footprint(self, request, pk=None):
        return self.retrieve(request, pk=pk)


class SpatialAssetViewSet(ModelViewSet):
    serializer_class = serializers.SpatialAssetSerializer
    queryset = models.SpatialAsset.objects.all()


class SignatureView(BaseRestViewMixin, views.APIView):
    """Generate an expirey URL signature."""

    def post(self, request):
        signer = UserSigner()
        signature = signer.sign(user=self.request.user)
        param = getattr(settings, 'RGD_SIGNED_URL_QUERY_PARAM', 'signature')
        return response.Response({param: signature})



from django.contrib.auth.models import User
from rest_framework.authtoken.models import Token
from rgd_client import create_rgd_client
from rgd_client.plugin import CorePlugin, RgdPlugin


# Plugin A
class PluginA(RgdPlugin):
    rgd = CorePlugin


class ClientA:
    a = PluginA


# Plugin B
class PluginB(RgdPlugin):
    a = PluginA
    rgd = CorePlugin


class ClientB:
    b = PluginB


class MyClient(ClientA, ClientB):
    pass


def test_plugin_dependency_inject(live_server):

    params = {'username': 'test@kitware.com', 'email': 'test@kitware.com', 'password': 'password'}

    user = User.objects.create_user(is_staff=True, is_superuser=True, **params)
    user.save()

    # use constant value for API key so this client fixture can be reused across multiple tests
    Token.objects.create(user=user, key='topsecretkey')

    client = create_rgd_client(
        username=params['username'],
        password=params['password'],
        api_url=f'{live_server.url}/api',
        extra_plugins=[ClientA, ClientB],
    )

    # Assert plugins were loaded correctly
    assert isinstance(client.a.rgd, CorePlugin)
    assert isinstance(client.b.rgd, CorePlugin)
    assert isinstance(client.b.a, PluginA)

    # Assert underlying class is not modified
    assert not isinstance(PluginA.rgd, CorePlugin)
    assert not isinstance(PluginB.rgd, CorePlugin)
    assert not isinstance(PluginB.a, PluginA)



from pathlib import Path

import pytest
from rest_framework import status
from rgd.models import ChecksumFile


@pytest.mark.django_db(transaction=True)
def test_swagger(admin_api_client):
    response = admin_api_client.get('/swagger/?format=openapi')
    assert status.is_success(response.status_code)


@pytest.mark.django_db(transaction=True)
def test_download_checksum_file(admin_api_client, checksum_file):
    pk = checksum_file.pk
    response = admin_api_client.get(f'/api/rgd/checksum_file/{pk}/data')
    assert status.is_redirect(response.status_code)


@pytest.mark.django_db(transaction=True)
def test_download_checksum_file_url(admin_api_client, checksum_file_url):
    pk = checksum_file_url.pk
    response = admin_api_client.get(f'/api/rgd/checksum_file/{pk}/data')

    assert response.status_code == 200
    assert response.headers['Content-Type'] == 'application/json'
    assert response.json() == checksum_file_url.url


@pytest.mark.django_db(transaction=True)
def test_get_checksum_file(admin_api_client, checksum_file):
    pk = checksum_file.pk
    content = admin_api_client.get(f'/api/rgd/checksum_file/{pk}').data
    assert content
    # Check that a hyperlink is given to the file data
    # NOTE: tried using the URLValidator from django but it thinks this URL is invalid
    assert content['file'].startswith('http')


@pytest.mark.django_db(transaction=True)
def test_get_spatial_entry(admin_api_client, spatial_asset_a):
    """Test individual GET for SpatialEntry model."""
    pk = spatial_asset_a.spatial_id
    response = admin_api_client.get(f'/api/rgd/spatial_entry/{pk}')
    assert response.status_code == 200
    assert response.data
    assert response.data['outline']


@pytest.mark.django_db(transaction=True)
def test_get_checksum_file_tree(
    checksum_file_factory, checksum_file_url: ChecksumFile, admin_api_client
):
    """Test that the tree list endpoint functions as expected."""
    # Top level
    file1: ChecksumFile = checksum_file_factory(name='h.txt')

    # a/ directory
    file2: ChecksumFile = checksum_file_factory(name='a/f.txt')
    file3: ChecksumFile = checksum_file_factory(name='a/g.txt')

    # a/b/ directory
    file4: ChecksumFile = checksum_file_factory(name='a/b/d.txt')
    file5: ChecksumFile = checksum_file_factory(name='a/b/e.txt')

    # Put one url file in a/b/ directory
    checksum_file_url.name = 'a/b/c.jpg'
    checksum_file_url.save()
    file6 = checksum_file_url

    # Testing folders
    folder_b = [file4, file5, file6]
    folder_a = [file2, file3] + folder_b

    known_size_a = sum([f.size for f in folder_a if f.size is not None])
    a_created = min([file.created for file in folder_a]).isoformat()
    a_modified = max([file.modified for file in folder_a]).isoformat()

    known_size_b = sum([f.size for f in folder_b if f.size is not None])
    b_created = min([file.created for file in folder_b]).isoformat()
    b_modified = max([file.modified for file in folder_b]).isoformat()

    # Check top level response
    r = admin_api_client.get('/api/rgd/checksum_file/tree')
    assert r.json()['files'][file1.name]['id'] == file1.id
    assert r.json()['folders'] == {
        'a': {
            'known_size': known_size_a,
            'num_files': len(folder_a),
            'num_url_files': 1,
            'created': a_created,
            'modified': a_modified,
        }
    }

    # Search folder 'a'
    r = admin_api_client.get('/api/rgd/checksum_file/tree', {'path_prefix': 'a'})

    # Files
    files = r.json()['files']
    assert files[Path(file2.name).name]['id'] == file2.id
    assert files[Path(file3.name).name]['id'] == file3.id

    # Folders
    assert r.json()['folders'] == {
        'b': {
            'known_size': known_size_b,
            'num_files': len(folder_b),
            'num_url_files': 1,
            'created': b_created,
            'modified': b_modified,
        }
    }

    # Search folder 'a/b'
    r = admin_api_client.get('/api/rgd/checksum_file/tree', {'path_prefix': 'a/b'})

    # Files
    files = r.json()['files']
    assert files[Path(file4.name).name]['id'] == file4.id
    assert files[Path(file5.name).name]['id'] == file5.id
    assert files[Path(file6.name).name]['id'] == file6.id



# Generated by Django 3.2.10 on 2021-12-15 06:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0006_auto_20211210_1744'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='checksumfile',
            name='unique_url_collection',
        ),
        migrations.AddConstraint(
            model_name='checksumfile',
            constraint=models.UniqueConstraint(
                condition=models.Q(('type', 2)),
                fields=('collection', 'url'),
                name='unique_url_collection',
            ),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-25 18:46

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('rgd', '0008_collection_description'),
    ]

    operations = [
        migrations.AlterField(
            model_name='checksumfile',
            name='collection',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='%(class)ss',
                related_query_name='%(class)ss',
                to='rgd.collection',
            ),
        ),
        migrations.AlterField(
            model_name='collectionpermission',
            name='collection',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to='rgd.collection'
            ),
        ),
        migrations.AlterField(
            model_name='collectionpermission',
            name='user',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AlterField(
            model_name='spatialasset',
            name='files',
            field=models.ManyToManyField(related_name='+', to='rgd.checksumfile'),
        ),
    ]



import pytest
from rest_framework.authtoken.views import ObtainAuthToken
from rgd import models
from rgd.permissions import filter_read_perm, filter_write_perm
from rgd.rest.mixins import BaseRestViewMixin
from rgd.urls import urlpatterns
from rgd.views import PermissionDetailView, PermissionListView, PermissionTemplateView


@pytest.mark.django_db(transaction=True)
def test_unassigned_permissions_complex(
    user_factory, user, spatial_asset_a, spatial_asset_b, settings
):
    # TODO: reimplement with multi raster file fixture also
    settings.RGD_GLOBAL_READ_ACCESS = False
    admin = user_factory(is_superuser=True)
    admin_q = filter_read_perm(admin, models.SpatialEntry.objects.all())
    basic_q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert len(admin_q) == 2
    assert len(basic_q) == 0
    admin_q = filter_read_perm(admin, models.Collection.objects.all())
    basic_q = filter_read_perm(user, models.Collection.objects.all())
    assert len(admin_q) == 2
    assert len(basic_q) == 0
    settings.RGD_GLOBAL_READ_ACCESS = True
    admin_q = filter_read_perm(admin, models.SpatialEntry.objects.all())
    basic_q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert len(admin_q) == len(basic_q)
    assert set(admin_q) == set(basic_q)
    admin_q = filter_read_perm(admin, models.Collection.objects.all())
    basic_q = filter_read_perm(user, models.Collection.objects.all())
    assert len(admin_q) == len(basic_q)
    assert set(admin_q) == set(basic_q)


@pytest.mark.django_db(transaction=True)
def test_nonadmin_user_permissions(user, spatial_asset_a, spatial_asset_b):
    # Filter and make sure nothing returns
    basic_q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert basic_q.count() == 0
    # Set collection on files and make sure both return
    collection = models.Collection.objects.create(name='test')
    permission = models.CollectionPermission.objects.create(
        collection=collection, user=user, role=models.CollectionPermission.READER
    )
    spatial_asset_a.files.update(collection=collection)
    spatial_asset_b.files.update(collection=collection)
    read_q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert read_q.count() == 2
    # Now make sure this user does not have write access
    write_q = filter_write_perm(user, models.SpatialEntry.objects.all())
    assert write_q.count() == 0
    # Update permissions to have write access and check
    permission.role = models.CollectionPermission.OWNER
    permission.save(
        update_fields=[
            'role',
        ]
    )
    write_q = filter_write_perm(user, models.SpatialEntry.objects.all())
    assert write_q.count() == 2


@pytest.mark.django_db(transaction=True)
def test_nonadmin_created_by_permissions(user, spatial_asset_a, spatial_asset_b):
    # Filter and make sure nothing returns
    q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert q.count() == 0
    # Update the `created_by` field and check that query works
    spatial_asset_a.files.update(created_by=user)
    spatial_asset_b.files.update(created_by=user)
    # NOTE: the ChecksumFileFactory sets the Collection by default
    q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert q.count() == 2
    # Update the `collection` field and check that query works
    spatial_asset_a.files.update(collection=None)
    spatial_asset_b.files.update(collection=None)
    q = filter_read_perm(user, models.SpatialEntry.objects.all())
    assert q.count() == 2


def test_urls():
    for pattern in urlpatterns:
        if hasattr(pattern.callback, 'view_class') and 'WrappedAPIView' not in str(
            pattern.callback
        ):
            assert issubclass(
                pattern.callback.view_class,
                (
                    BaseRestViewMixin,
                    PermissionDetailView,
                    PermissionTemplateView,
                    PermissionListView,
                    ObtainAuthToken,
                ),
            )



from django.conf import settings
from django.db import models


class Collection(models.Model):

    name = models.CharField(max_length=127, unique=True)
    description = models.TextField(blank=True, null=True)

    def __str__(self):
        return f'{self.name} ({self.pk})'

    class Meta:
        default_related_name = 'collections'

    def __len__(self):
        return self.checksumfiles.count()


class CollectionPermission(models.Model):
    READER = 1
    OWNER = 2
    ROLE_CHOICES = [
        (READER, 'Reader'),
        (OWNER, 'Owner'),
    ]

    collection = models.ForeignKey(Collection, on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    role = models.SmallIntegerField(
        choices=ROLE_CHOICES,
        default=READER,
        db_index=True,
        help_text=(
            'A "reader" can view assets in this collection. '
            'An "owner" can additionally add/remove other users, set their permissions, delete the collection, and add/remove other files.'
        ),
    )

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=['collection', 'user'],
                name='unique_user',
            )
        ]
        default_related_name = 'collection_permissions'



# Generated by Django 3.2.6 on 2021-08-20 15:56

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('rgd', '0002_alter_spatialasset_files'),
    ]

    operations = [
        migrations.AddField(
            model_name='checksumfile',
            name='created_by',
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from django.contrib.gis.db import models
from django_extensions.db.models import TimeStampedModel
from rgd.utility import get_cache_dir


class FileSet(TimeStampedModel):

    name = models.CharField(max_length=1000, blank=True, null=True)
    description = models.TextField(null=True, blank=True)

    @property
    def files(self):
        return self.checksumfile_set.all()

    def get_cache_path(self):
        """Generate a predetermined path in the cache directory.

        This makes a directory under the cache directory for each ChecksumFile
        record. The subdirectory is the Primary Key of the record and the nested
        files are the file itself, lockfile, and any additional files used or
        generated by third party libraries.

        `<cache directory>/<pk>/*`

        """
        p = get_cache_dir() / f'{self.pk}'
        p.mkdir(exist_ok=True, parents=True)
        return p

    def yield_all_to_local_path(self, directory: str = None):
        """Download all the files in this file set to a local path.

        Please note that this uses a contextmanager to acquire a lock on the
        directory to make sure the files are not automatically cleaned up by
        other threads or processes.

        """
        from .utils import yield_checksumfiles  # avoiding circular import

        if not directory:
            directory = self.get_cache_path()
        return yield_checksumfiles(self.files, directory)



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.models import User
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.models import SpatialAsset, WhitelistedEmail

from .mixins import MODIFIABLE_FILTERS, _FileGetNameMixin


def make_users_active(modeladmin, request, queryset):
    """Make each user active."""
    for user in queryset.all():
        if not user.is_active:
            user.is_active = True
            user.save(update_fields=['is_active'])


def make_users_staff(modeladmin, request, queryset):
    """Make each user staff."""
    for user in queryset.all():
        if not user.is_staff:
            user.is_staff = True
            user.save(update_fields=['is_staff'])


admin.site.unregister(User)


@admin.register(User)
class CustomUserAdmin(UserAdmin):
    actions = (
        make_users_active,
        make_users_staff,
    )

    list_display = (
        'username',
        'is_staff',
        'is_superuser',
        'is_active',
    )


@admin.register(WhitelistedEmail)
class WhitelistedEmailAdmin(OSMGeoAdmin):
    list_display = (
        'pk',
        'email',
    )


@admin.register(SpatialAsset)
class SpatialAssetAdmin(OSMGeoAdmin, _FileGetNameMixin):
    list_display = (
        'pk',
        'get_name',
        'modified',
        'created',
    )
    list_filter = MODIFIABLE_FILTERS
    raw_id_fields = ('files',)



import json
import os

from django.contrib.gis.db import models
from django_extensions.db.models import TimeStampedModel
from model_utils.managers import InheritanceManager

from .constants import DB_SRID
from .file import ChecksumFile


class SpatialEntry(models.Model):
    """Common model to all geospatial data entries.

    This is intended to be used in a mixin manner.

    """

    # `InheritanceManager` allows us to select inhereted tables via `objects.select_subclasses()`
    objects = InheritanceManager()

    spatial_id = models.AutoField(primary_key=True)

    # Datetime of creation for the dataset
    acquisition_date = models.DateTimeField(null=True, default=None, blank=True)

    # This can be used with GeoDjango's geographic database functions for spatial indexing
    footprint = models.GeometryField(srid=DB_SRID)
    outline = models.GeometryField(srid=DB_SRID)

    instrumentation = models.CharField(
        max_length=100,
        null=True,
        blank=True,
        help_text='The instrumentation used to acquire these data.',
    )

    def __str__(self):
        try:
            return 'Spatial ID: {} (ID: {}, type: {})'.format(self.spatial_id, self.id, type(self))
        except AttributeError:
            return super().__str__()

    @property
    def bounds(self):
        extent = {
            'xmin': self.outline.extent[0],
            'ymin': self.outline.extent[1],
            'xmax': self.outline.extent[2],
            'ymax': self.outline.extent[3],
        }
        return extent

    @property
    def bounds_json(self):
        return json.dumps(self.bounds)

    @property
    def basename(self):
        return os.path.basename(self.name)


class WhitelistedEmail(models.Model):
    """Pre-approve users for sign up by their email."""

    email = models.EmailField()


class SpatialAsset(SpatialEntry, TimeStampedModel):
    """Any spatially referenced file set.

    This can be any collection of files that have a spatial reference and are
    not explicitly handled by the other SpatialEntry subtypes. For example, this
    model can be used to hold a collection of PDF documents or slide decks that
    have a georeference.

    """

    files = models.ManyToManyField(ChecksumFile, related_name='+')



from django.contrib.gis import forms
from django.contrib.gis.db.models.functions import Area, GeometryDistance, Intersection
from django.contrib.gis.measure import D
from django.core.validators import RegexValidator
from django.db.models import F, FloatField, Q
from django.db.models.functions import Cast
from django_filters import rest_framework as filters
from rgd.models import ChecksumFile, Collection, SpatialEntry
from rgd.permissions import get_paths


class GeometryFilter(filters.Filter):
    field_class = forms.GeometryField
    # Ensures GeoJSON objects are converted to correct SRID
    field_class.widget.map_srid = 4326


class SpatialEntryFilter(filters.FilterSet):

    q = GeometryFilter(
        help_text='A Well-known text (WKT) representation of a geometry or a GeoJSON.',
        label='WKT/GeoJSON',
        method='filter_q',
    )
    predicate = filters.ChoiceFilter(
        choices=(
            ('contains', 'contains'),
            ('crosses', 'crosses'),
            ('disjoint', 'disjoint'),
            ('equals', 'equals'),
            ('intersects', 'intersects'),
            ('overlaps', 'overlaps'),
            ('touches', 'touches'),
            ('within', 'within'),
        ),
        help_text=(
            'A named spatial predicate based on the DE-9IM. This spatial predicate will be used '
            'to filter data such that `predicate(a, b)` where `b` is the queried geometry.'
        ),
        label='Spatial predicate',
        method='filter_predicate',
    )
    relates = filters.CharFilter(
        help_text=(
            'Specify exactly how the queried geometry should relate to the data using a DE-9IM '
            'string code.'
        ),
        label='DE-9IM string code',
        max_length=9,
        method='filter_relates',
        min_length=9,
        validators=(
            RegexValidator(regex=r'^[\*012TF]{9}$', message='Enter a valid DE-9IM string code.'),
        ),
    )
    distance = filters.RangeFilter(
        help_text='The minimum/maximum distance around the queried geometry in meters.',
        label='Distance',
        method='filter_distance',
    )
    acquired = filters.IsoDateTimeFromToRangeFilter(
        field_name='acquisition_date',
        help_text='The ISO 8601 formatted date and time when data was acquired.',
        label='Acquired',
    )
    instrumentation = filters.CharFilter(
        field_name='instrumentation',
        help_text='The instrumentation used to acquire at least one of these data.',
        label='Instrumentation',
        lookup_expr='icontains',
    )
    time_of_day = filters.TimeRangeFilter(
        help_text='The minimum/maximum times during the day the records were acquired.',
        label='Time of Day',
        method='filter_time_of_day',
    )
    collections = filters.ModelMultipleChoiceFilter(
        help_text='One or more collections that the data might belong to.',
        label='Collections',
        method='filter_collection',
        queryset=Collection.objects.all(),
    )
    percent_overlap = filters.NumberFilter(
        help_text='The minute percent overlap with search geometry (between 0 and 1).',
        label='Percent overlap',
        method='filter_percent_overlap',
    )

    @property
    def _geometry(self):
        return self.form.cleaned_data['q']

    @property
    def _has_geom(self):
        return self._geometry is not None

    def filter_q(self, queryset, name, value):
        """Sort the queryset by distance to queried geometry.

        Annotates the queryset with `distance`.

        This uses the efficient KNN operation:
        https://postgis.net/docs/geometry_distance_knn.html
        """
        if value:
            queryset = queryset.annotate(distance=GeometryDistance('footprint', value)).order_by(
                'distance'
            )
        return queryset

    def filter_predicate(self, queryset, name, value):
        """Filter the spatial entries by the chosen predicate."""
        if value and self._has_geom:
            queryset = queryset.filter(**{f'footprint__{value}': self._geometry})
        return queryset

    def filter_relates(self, queryset, name, value):
        """Filter the spatial entries by the chosen DE-9IM."""
        if value and self._has_geom:
            queryset = queryset.filter(footprint__relates=(self._geometry, value))
        return queryset

    def filter_distance(self, queryset, name, value):
        """Filter the queryset by distance to the queried geometry.

        We may wish to use the distance in degrees later on. This is
        very taxing on the DBMS right now. The distance in degrees
        can be provided by the initial geometry query.
        """
        if value and self._has_geom:
            geom = self._geometry
            if value.start is not None:
                queryset = queryset.filter(footprint__distance_gte=(geom, D(m=value.start)))
            if value.stop is not None:
                queryset = queryset.filter(footprint__distance_lte=(geom, D(m=value.stop)))
        return queryset

    def filter_time_of_day(self, queryset, name, value):
        """Filter the queryset by time of day acquired.

        Use case: find all rasters acquired between 8am and 4pm
        for all days in the acquired date range (i.e. only daytime imagery)
        """
        if value is not None:
            queryset = queryset.annotate(time_of_day=F('acquisition_date__time'))
            if value.start is not None:
                queryset = queryset.filter(time_of_day__gte=value.start)
            if value.stop is not None:
                queryset = queryset.filter(time_of_day__lte=value.stop)
        return queryset

    def filter_collection(self, queryset, name, value):
        """Filter the queryset by the collection it belongs to."""
        if value:
            conditions = Q()
            for path in get_paths(queryset.model, ChecksumFile):
                conditions |= path.q(collection__in=value)
            return queryset.filter(conditions).distinct()
        return queryset

    def filter_percent_overlap(self, queryset, name, value: float):
        """Filter the queryset by percent overlap with the queried geometry."""
        if value is not None and value > 0 and value <= 1 and self._has_geom:
            geom = self._geometry
            queryset = queryset.filter(footprint__overlaps=geom).annotate(
                overlap_percentage=(
                    Cast(Area(Intersection(F('footprint'), geom)) / Area(geom), FloatField())
                )
            )
            queryset = queryset.filter(overlap_percentage__gte=value)
        return queryset

    class Meta:
        model = SpatialEntry
        fields = [
            'q',
            'predicate',
            'relates',
            'distance',
            'acquired',
            'percent_overlap',
            'instrumentation',
            'time_of_day',
            'collections',
        ]


class CollectionFilter(filters.FilterSet):
    name = filters.CharFilter(
        field_name='name',
        help_text='The name of the Collection.',
        label='Name',
        lookup_expr='exact',
    )

    class Meta:
        model = Collection
        fields = [
            'name',
        ]



# Generated by Django 3.2.10 on 2021-12-10 17:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0005_auto_20211105_1715'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='name',
            field=models.CharField(max_length=127, unique=True),
        ),
        migrations.AddConstraint(
            model_name='checksumfile',
            constraint=models.UniqueConstraint(
                fields=('collection', 'url'), name='unique_url_collection'
            ),
        ),
    ]



from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin
from rgd.models import Collection, CollectionPermission


class CollectionPermissionInline(admin.TabularInline):
    model = CollectionPermission
    fk_name = 'collection'
    fields = ('user', 'role')
    extra = 1


@admin.register(Collection)
class CollectionAdmin(OSMGeoAdmin):
    fields = ('name',)
    inlines = (CollectionPermissionInline,)



from collections import deque
import dataclasses
from dataclasses import dataclass
from itertools import chain
from typing import Any, Deque, Iterator, Optional, Type, Union

from django.conf import settings
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import User
from django.core.exceptions import PermissionDenied
from django.db.models import Model, Q, QuerySet
from django.db.models.fields.related import OneToOneRel, RelatedField
from rest_framework import filters, permissions
from rgd import models
from typing_extensions import TypeGuard

PathField = Union[RelatedField, OneToOneRel]


def is_path_field(field: Any) -> TypeGuard[PathField]:
    """Check if a `field` is a `PathField`."""
    is_parent = isinstance(field, OneToOneRel) and field.parent_link
    is_forward = isinstance(field, RelatedField)
    return is_parent or is_forward


def get_related_fields(model: Type[Model]) -> Iterator[PathField]:
    """Get all `PathField`s for a `model`."""
    for field in model._meta.get_fields():
        if is_path_field(field):
            yield field


def get_related_field(model: Type[Model], field_name: str) -> PathField:
    """Get a `PathField` for a `model` by its field name."""
    field = model._meta.get_field(field_name)
    if is_path_field(field):
        return field
    raise ValueError(f'{field} is not a `PathField`')


@dataclass
class IdentityPathField:
    """A self referential field used to represent a root node."""

    model: Type[Model]

    @property
    def related_model(self) -> Type[Model]:
        return self.model


@dataclass
class Path:
    """Singly linked list of `PathField`s."""

    field: Union[PathField, IdentityPathField]
    previous: Optional['Path'] = dataclasses.field(default=None, init=False)

    def traverse(self) -> Iterator['Path']:
        """Iterate backwards over the linked paths."""
        path: Optional['Path'] = self
        while path is not None:
            yield path
            path = path.previous

    def has_visited(self, model: Type[Model]) -> bool:
        """Return whether this path has visited the given `model`."""
        return any(
            path.field.related_model == model or path.field.model == model
            for path in self.traverse()
        )

    def get_frontier(self) -> Iterator['Path']:
        """Iterate over the paths in this path's frontier."""
        for field in get_related_fields(self.field.related_model):
            if not self.has_visited(field.related_model):
                path = Path(field)
                path.previous = self
                yield path

    def q(self, **kwargs: Any) -> Q:
        """Return a `Q` object for this path.

        Can be any keyword arguments i.e. `user__isnull=True`.
        """
        conditions = Q()
        field_names = reversed(
            tuple(
                path.field.name
                for path in self.traverse()
                if not isinstance(path.field, IdentityPathField)
            )
        )
        for key, value in kwargs.items():
            dunder = '__'.join(chain(field_names, (key,)))
            conditions &= Q(**{dunder: value})
        return conditions


def get_paths(source: Type[Model], target: Type[Model]) -> Iterator[Path]:
    """Get all paths from a `source` model to a `target` model.

    Cycles are prevented by not following models that have already been
    visited. This means that recursive relationships will only be followed
    to a depth of 1.

    Only "forward" relationships are followed. This includes
    parent -> child relationships for inherited models.

    This performs a breadth-first search, so shorter paths will be
    returned first.
    """
    queue: Deque[Path] = deque()
    queue.append(Path(IdentityPathField(source)))

    while queue:
        path = queue.pop()
        if path.field.related_model == target:
            yield path
            continue
        frontier = path.get_frontier()
        queue.extendleft(frontier)


def filter_perm(user, queryset, role):
    """Filter a queryset.

    Main authorization business logic goes here.
    """
    # Called outside of view
    if user is None:
        # TODO: I think this is used if a user isn't logged in and hits our endpoints which is a problem
        return queryset
    # Must be logged in
    if not user.is_active or user.is_anonymous:
        return queryset.none()
    # Superusers can see all (not staff users)
    if user.is_active and user.is_superuser:
        return queryset
    # Check permissions
    conditions = []
    model = queryset.model
    paths_to_checksumfile = [*get_paths(model, models.ChecksumFile)]
    if model == models.Collection:
        # Add custom reverse relationships
        field = model._meta.get_field('checksumfiles')
        path = Path(field)
        paths_to_checksumfile.append(path)
    for path in paths_to_checksumfile:
        # A user can read/write a file if they are the creator
        is_creator = path.q(created_by=user)
        conditions.append(is_creator)
        if (
            getattr(settings, 'RGD_GLOBAL_READ_ACCESS', False)
            and role == models.CollectionPermission.READER
        ):
            # A user can read any file by default
            has_no_owner = path.q(created_by__isnull=True)
            conditions.append(has_no_owner)
    for path in get_paths(model, models.Collection):
        # Check collection permissions
        has_permission = path.q(collection_permissions__user=user)
        has_role_level = path.q(collection_permissions__role__gte=role)
        conditions.append(has_permission & has_role_level)
    whitelist = (
        queryset.none()
        .union(*(queryset.filter(condition) for condition in conditions), all=True)
        .values('pk')
    )
    return queryset.filter(pk__in=whitelist)


def filter_read_perm(user, queryset):
    """Filter a queryset to what the user may read."""
    return filter_perm(user, queryset, models.CollectionPermission.READER)


def filter_write_perm(user: User, queryset: QuerySet):
    """Filter a queryset to what the user may edit."""
    return filter_perm(user, queryset, models.CollectionPermission.OWNER)


def check_read_perm(user: User, obj: Model):
    """Raise 'PermissionDenied' error if user does not have read permissions."""
    model = type(obj)
    if not filter_read_perm(user, model.objects.filter(pk=obj.pk)).exists():
        raise PermissionDenied


def check_write_perm(user: User, obj: Model):
    """Raise 'PermissionDenied' error if user does not have write permissions."""
    # Called outside of view
    model = type(obj)
    if not filter_write_perm(user, model.objects.filter(pk=obj.pk)).exists():
        raise PermissionDenied


class CollectionAuthorizationBackend(BaseBackend):
    def has_perm(self, user, perm, obj=None):
        """Supplement default Django permission backend.

        Returns `True` if the user has the specified permission, where perm is in the format
        `"<app label>.<permission codename>"`. If the user is
        inactive, this method will always return False. For an active superuser, this method
        will always return `True`.

        https://docs.djangoproject.com/en/3.1/ref/contrib/auth/#django.contrib.auth.models.User.has_perm
        """
        app_label, codename = perm.split('.')
        if app_label == 'geodata':
            if codename.startswith('view'):
                check_read_perm(user, obj)
            if (
                codename.startswith('add')
                or codename.startswith('delete')
                or codename.startswith('change')
            ):
                check_write_perm(user, obj)


class CollectionAuthorization(permissions.BasePermission):
    """A custom Django REST Framework permission backend.

    Object-level permission to only allow owners of an object to edit it
    and readers to view it. This will not filter the queryset in 'list' views.
    You should use this in combination with the `CollectionAuthorizationFilter` to
    also filter the queryset in 'list' views.

    https://www.django-rest-framework.org/api-guide/permissions/#custom-permissions
    """

    def has_object_permission(self, request, view, obj):
        # Check read permissions for GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            check_read_perm(request.user, obj)
        else:
            check_write_perm(request.user, obj)
        # The above calls should raise an exception if the user does not have permissions.
        return True


class CollectionAuthorizationFilter(filters.BaseFilterBackend):
    """A custom Django REST Framework filter backend for permissions.

    As a result of applying this filter, objects that the user does not have persmissions
    to view will return a '404' since the queryset is filtered before object-level
    permissions are checked.

    https://www.django-rest-framework.org/api-guide/permissions/#custom-permissions
    """

    def filter_queryset(self, request, queryset, view):
        # Check read permissions for GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return filter_read_perm(request.user, queryset)
        # Otherwise, assume this is a writable query.
        return filter_write_perm(request.user, queryset)



# Generated by Django 3.2.5 on 2021-08-12 11:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='spatialasset',
            name='files',
            field=models.ManyToManyField(
                related_name='_rgd_spatialasset_files_+', to='rgd.ChecksumFile'
            ),
        ),
    ]



import json
from pathlib import Path
import re
from typing import Optional

import pytest
from requests.exceptions import HTTPError
from rgd.models import ChecksumFile, Collection
from rgd_client import RgdClient, create_rgd_client
from rgd_client.utils import API_KEY_DIR_PATH, API_KEY_FILE_NAME


@pytest.mark.django_db(transaction=True)
def test_basic_search(py_client: RgdClient, spatial_asset_a):
    bbox = {
        'type': 'Polygon',
        'coordinates': [
            [
                [-84.10240173339842, 39.72989765591686],
                [-84.0395736694336, 39.72989765591686],
                [-84.0395736694336, 39.78057429170315],
                [-84.10240173339842, 39.78057429170315],
                [-84.10240173339842, 39.72989765591686],
            ]
        ],
    }

    q = py_client.rgd.search(query=json.dumps(bbox), predicate='intersects')
    assert q['count'] == 1
    assert len(q['results']) == 1


@pytest.mark.django_db(transaction=True)
def test_create_file_from_url(py_client: RgdClient, checksum_file_url: ChecksumFile):
    """Test that creation of a ChecksumFile with a URL succeeds."""
    file_dict = py_client.rgd.create_file_from_url(checksum_file_url.get_url())
    assert file_dict['url'] == checksum_file_url.get_url()
    assert file_dict['type'] == 2


@pytest.mark.django_db(transaction=True)
def test_create_file_from_url_existing(
    py_client: RgdClient, checksum_file_url: ChecksumFile, collection: Collection
):
    """Test that create with an existing URL/Collection returns the existing file."""
    checksum_file_url.collection = collection
    checksum_file_url.save()

    file_dict = py_client.rgd.create_file_from_url(
        url=checksum_file_url.get_url(), collection=collection.id
    )
    assert file_dict['id'] == checksum_file_url.id


@pytest.mark.django_db(transaction=True)
def test_tree_file_search(py_client: RgdClient, checksum_file: ChecksumFile):
    """Test that basic connection to the tree endpoint succeeds."""
    resp_dict = py_client.rgd.file_tree_search()
    assert 'folders' in resp_dict
    assert 'files' in resp_dict


@pytest.mark.django_db(transaction=True)
def test_save_api_key(live_server, user_with_api_key):
    """Test that saving an API key works correctly."""
    username, password, api_token = user_with_api_key

    create_rgd_client(
        username=username,
        password=password,
        api_url=f'{live_server.url}/api',
        save=True,  # save the API key
    )

    # Ensure ~/.rgd/token exists and contains the user's API key
    assert (API_KEY_DIR_PATH / API_KEY_FILE_NAME).read_text() == api_token


@pytest.mark.django_db(transaction=True)
def test_invalid_local_api_key(live_server, user_with_api_key):
    """Test that a new API key is fetched when the one on disk is invalid."""
    username, password, api_token = user_with_api_key

    # Save invalid API key to disk so `create_rgd_client` attempts to use it
    (API_KEY_DIR_PATH / API_KEY_FILE_NAME).write_text(f'{api_token}_bad')

    create_rgd_client(
        username=username,
        password=password,
        api_url=f'{live_server.url}/api',
        save=True,  # save the API key
    )

    # Ensure that `create_rgd_client` overwrote the invalid key with a fresh one
    assert (API_KEY_DIR_PATH / API_KEY_FILE_NAME).read_text() == api_token


@pytest.mark.django_db(transaction=True)
def test_invalid_env_var_api_key(live_server, user_with_api_key, monkeypatch):
    """Test that a new API key is *not* fetched when one provided via env var is invalid."""
    username, password, api_token = user_with_api_key
    monkeypatch.setenv('RGD_API_TOKEN', api_token)

    # Make sure ~/.rgd/token doesn't exist
    (API_KEY_DIR_PATH / API_KEY_FILE_NAME).unlink(missing_ok=True)
    assert not (API_KEY_DIR_PATH / API_KEY_FILE_NAME).exists()

    create_rgd_client(
        username=username,
        password=password,
        api_url=f'{live_server.url}/api',
        save=True,  # save the API key
    )

    # Make sure ~/.rgd/token wasn't created
    assert not (API_KEY_DIR_PATH / API_KEY_FILE_NAME).exists()


@pytest.mark.django_db(transaction=True)
def test_api_url_trailing_slash(live_server, user_with_api_key, checksum_file_url: ChecksumFile):
    url: str = live_server.url

    # Ensure url doesn't end with a trailing slash already
    if url.endswith('/'):
        url = url.rstrip('/')

    username, password, _ = user_with_api_key

    py_client = create_rgd_client(
        username=username,
        password=password,
        api_url=f'{url}/api/',  # append a trailing slash to url
    )
    assert py_client.session.base_url == f'{url}/api/'

    # Make sure the client behaves as expected given a trailing slash
    file_dict = py_client.rgd.create_file_from_url(checksum_file_url.get_url())
    assert file_dict['url'] == checksum_file_url.get_url()


@pytest.mark.django_db(transaction=True)
def test_invalid_api_url(live_server, user_with_api_key, monkeypatch):
    """Test that the client fails loudly when given an invalid API URL."""
    url: str = live_server.url
    username, password, api_key = user_with_api_key

    monkeypatch.setenv('RGD_API_TOKEN', api_key)

    with pytest.raises(HTTPError):
        create_rgd_client(
            username=username,
            password=password,
            api_url=f'{url}/invalid',
        )


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize(
    # Test cases: when download_location is None, and when it is a valid string
    'download_path,expected_download_directory,use_id',
    [
        (None, Path.cwd(), False),
        (Path('~/.rgd/temp').expanduser(), Path('~/.rgd/temp').expanduser(), False),
        (None, Path.cwd(), True),
        (Path('~/.rgd/temp').expanduser(), Path('~/.rgd/temp').expanduser(), True),
    ],
)
def test_file_download(
    py_client: RgdClient,
    checksum_file: ChecksumFile,
    download_path: Optional[Path],
    expected_download_directory: Optional[Path],
    use_id: bool,
):
    expected_download_path: Path = expected_download_directory / (
        str(checksum_file.id) if use_id else checksum_file.name
    )

    # Remove file if it already exists
    expected_download_path.unlink(missing_ok=True)

    # Download file and make sure it exists and the contents are as expected.
    download_path = py_client.rgd.download_checksum_file_to_path(
        checksum_file.id, download_path, use_id=use_id
    )

    assert download_path == expected_download_path
    assert expected_download_path.read_bytes() == checksum_file.file.file.read()


@pytest.mark.django_db(transaction=True)
def test_invalid_file_download(py_client: RgdClient, checksum_file: ChecksumFile):
    """Test that attempting to download a non-existent file fails properly."""
    checksum_file.delete()

    with pytest.raises(HTTPError) as error:
        py_client.rgd.download_checksum_file_to_path(checksum_file.id)

    # Make sure it was a 404 error
    assert error.match(r'404 Client Error')


@pytest.mark.django_db(transaction=True)
def test_file_download_keep_existing(
    tmp_path: Path, py_client: RgdClient, checksum_file: ChecksumFile
):
    # Save a file where the client would normally put the downloaded file
    file_path = tmp_path / str(checksum_file.id)
    existing_file_content: bytes = b'foobar'

    with open(file_path, 'wb') as f:
        f.write(existing_file_content)

    # Make sure the file isn't modified
    assert existing_file_content == file_path.read_bytes()
    py_client.rgd.download_checksum_file_to_path(
        checksum_file.id, tmp_path, keep_existing=True, use_id=True
    )
    assert existing_file_content == file_path.read_bytes()

    # Make sure the file *is* modified
    py_client.rgd.download_checksum_file_to_path(
        checksum_file.id, tmp_path, keep_existing=False, use_id=True
    )
    assert existing_file_content != file_path.read_bytes()


@pytest.mark.django_db(transaction=True)
def test_create_collection_new(py_client: RgdClient):
    """Test that creating a new collection succeeds."""
    assert Collection.objects.count() == 0

    r = py_client.rgd.create_collection('test')
    assert r == {
        'id': Collection.objects.order_by('-id').first().id,
        'name': 'test',
        'description': None,
        'len': 0,
    }

    assert Collection.objects.count() == 1


@pytest.mark.django_db(transaction=True)
def test_create_collection_existing(py_client: RgdClient, collection: Collection):
    """Test that attempting to create an existing collection simply returns the existing one."""
    collection_count = Collection.objects.count()

    r = py_client.rgd.create_collection(collection.name)

    assert r == {
        'id': collection.id,
        'name': collection.name,
        'description': collection.description,
        'len': len(collection),
    }

    # Ensure that no new collections have been created
    assert Collection.objects.count() == collection_count


@pytest.mark.django_db(transaction=True)
def test_get_collection_by_id(py_client: RgdClient, collection: Collection):
    r = py_client.rgd.get_collection(collection.id)

    assert r == {
        'id': collection.id,
        'name': collection.name,
        'description': collection.description,
        'len': len(collection),
    }


@pytest.mark.django_db(transaction=True)
def test_get_collection_item(
    py_client: RgdClient, checksum_file_url: ChecksumFile, collection: Collection
):
    checksum_file_url.collection = collection
    checksum_file_url.save()

    item = py_client.rgd.get_collection_item(collection.id, 0)
    assert item['id'] == checksum_file_url.id
    assert item['name'] == checksum_file_url.name


@pytest.mark.django_db(transaction=True)
def test_get_collection_by_name_exists(py_client: RgdClient, collection: Collection):
    r = py_client.rgd.get_collection_by_name(collection.name)

    assert r == {
        'id': collection.id,
        'name': collection.name,
        'description': collection.description,
        'len': len(collection),
    }


@pytest.mark.django_db(transaction=True)
def test_get_collection_by_name_nonexistant(py_client: RgdClient, collection: Collection):
    nonexistant_collection_name = f'not_{collection.name}'

    with pytest.raises(ValueError) as error:
        py_client.rgd.get_collection_by_name(nonexistant_collection_name)

    assert error.match(re.escape(rf'Collection ({nonexistant_collection_name}) cannot be found.'))



import multiprocessing
import re
from typing import Generator, Optional

import boto3
import botocore
import djclick as click

from . import _data_helper as helper


def _iter_matching_objects(
    s3_client,
    bucket: str,
    prefix: str,
    include_regex: str,
    exclude_regex: str,
) -> Generator[dict, None, None]:
    paginator = s3_client.get_paginator('list_objects_v2')
    page_iter = paginator.paginate(Bucket=bucket, Prefix=prefix)
    include_pattern = re.compile(include_regex)
    exclude_pattern = re.compile(exclude_regex)

    for page in page_iter:
        for obj in page['Contents']:
            if include_pattern.match(obj['Key']) and not exclude_pattern.search(obj['Key']):
                yield obj


class Loader:
    def __init__(self, bucket: str, region: str, google: bool = False):
        self.bucket = bucket
        self.google = google
        self.region = region

    def _format_url(self, base_url):
        if self.google:
            return f'http://storage.googleapis.com/{base_url}'
        return f'https://{self.region}.amazonaws.com/{base_url}'

    def load_object(self, obj: dict) -> None:
        key = obj['Key']
        url = self._format_url(f'{self.bucket}/{key}')
        helper._get_or_create_checksum_file_url(url, name=key)


@click.command()
@click.argument('bucket')
@click.option('--include-regex', default='')
@click.option('--exclude-regex', default='')
@click.option('--prefix', default='')
@click.option('--region', default='us-east-1')
@click.option('--access-key-id')
@click.option('--secret-access-key')
@click.option('--google', is_flag=True, default=False)
def ingest_s3(
    bucket: str,
    include_regex: str,
    exclude_regex: str,
    prefix: str,
    region: str,
    access_key_id: Optional[str],
    secret_access_key: Optional[str],
    google: bool,
) -> None:
    if access_key_id and secret_access_key:
        boto3_params = {
            'aws_access_key_id': access_key_id,
            'aws_secret_access_key': secret_access_key,
            'config': botocore.client.Config(signature_version='s3v4', region_name=region),
        }
    else:
        boto3_params = {
            'config': botocore.client.Config(
                signature_version=botocore.UNSIGNED, region_name=region
            )
        }

    if google:  # Google Cloud Storage
        boto3_params['endpoint_url'] = 'https://storage.googleapis.com'

    s3_client = boto3.client('s3', **boto3_params)

    loader = Loader(bucket, region, google=google)
    pool = multiprocessing.Pool(multiprocessing.cpu_count())
    pool.map(
        loader.load_object,
        _iter_matching_objects(s3_client, bucket, prefix, include_regex, exclude_regex),
    )



from rest_framework import mixins, viewsets
from rgd.rest.mixins import BaseRestViewMixin


class ReadOnlyModelViewSet(
    BaseRestViewMixin,
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """A viewset for read models.

    Provides default `list()` and `retrieve()` actions.
    """

    pass


class ModelViewSet(
    BaseRestViewMixin,
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """
    A viewset for read and edit models.

    Provides default `create()`, `retrieve()`, `update()`,
    `partial_update()`, `destroy()` and `list()` actions.
    """

    pass



from rgd import datastore


def test_datastore_format():
    """Validates that all entries in the datastore are formatted correctly."""
    for k, v in datastore.registry.items():
        c = v.split(':')
        assert len(c) == 2, f'`{k}` improperly formatted.'


def test_datastore_valid():
    """Validates that all entries in the datastore are available for download."""
    for k in datastore.registry.keys():
        assert datastore.datastore.is_available(k), f'`{k}` not available.'



import os

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.core.validators import URLValidator
from rgd import models
from rgd.datastore import datastore
from rgd.utility import get_or_create_no_commit


class SynchronousTasksCommand(BaseCommand):
    def set_synchronous(self):
        # Set celery to run all tasks synchronously
        self._eager = getattr(settings, 'CELERY_TASK_ALWAYS_EAGER', False)
        self._prop = getattr(settings, 'CELERY_TASK_EAGER_PROPAGATES', False)
        settings.CELERY_TASK_ALWAYS_EAGER = True
        settings.CELERY_TASK_EAGER_PROPAGATES = True

    def reset_celery(self):
        # Reset celery to previous settings
        settings.CELERY_TASK_ALWAYS_EAGER = self._eager
        settings.CELERY_TASK_EAGER_PROPAGATES = self._prop


def _save_signal(entry, created):
    if not created and entry.status == models.mixins.Status.SUCCEEDED:
        entry.skip_signal = True
    entry.save()


def _get_or_create_checksum_file_url(url: str, name=None):
    URLValidator()(url)  # raises `ValidationError` if not a valid URL
    try:
        file_entry = models.ChecksumFile.objects.get(url=url)
        _save_signal(file_entry, False)
        if name:
            file_entry.name = name
            file_entry.save(update_fields=['name'])
    except models.ChecksumFile.DoesNotExist:
        file_entry = models.ChecksumFile()
        file_entry.url = url
        file_entry.type = models.FileSourceType.URL
        if not name:
            # this is to prevent calling `urlopen` in the save to get the file name.
            # this is not a great way to set the default name, but its fast
            name = os.path.basename(url)
        file_entry.name = name
        _save_signal(file_entry, True)
    return file_entry


def _get_or_create_checksum_file_filefield(file: str, name=None, use_datastore: bool = True):
    try:
        if name is None:
            file_entry = models.ChecksumFile.objects.get(name=file)
        else:
            file_entry = models.ChecksumFile.objects.get(name=name)
        _save_signal(file_entry, False)
    except models.ChecksumFile.DoesNotExist:
        if use_datastore:
            path = datastore.fetch(file)
        else:
            path = file
        file_entry = models.ChecksumFile()
        if name:
            file_entry.name = name
        else:
            file_entry.name = file
        with open(path, 'rb') as f:
            file_entry.save_file_contents(f, os.path.basename(path))
        file_entry.type = models.FileSourceType.FILE_FIELD
        _save_signal(file_entry, True)
    return file_entry


def _get_or_create_checksum_file(file: str, name=None):
    # Check if there is already an image file with this URL or name
    #  to avoid duplicating data
    try:
        file_entry = _get_or_create_checksum_file_url(file, name=name)
    except ValidationError:
        try:
            file_entry = _get_or_create_checksum_file_filefield(file, name=name, use_datastore=True)
        except ValueError:
            file_entry = _get_or_create_checksum_file_filefield(
                file, name=name, use_datastore=False
            )
    return file_entry


def _get_or_create_file_model(model, file, name=None):
    # For models that point to a `ChecksumFile`
    file_entry = _get_or_create_checksum_file(file, name=name)
    # No commit in case we need to skip the signal
    entry, created = get_or_create_no_commit(model, file=file_entry)
    _save_signal(entry, created)
    return entry



"""Mixin helper classes."""
from typing import Iterable

from celery import Task
from django.conf import settings
from django.contrib.gis.db import models
from django.utils.translation import gettext_lazy as _


class Status(models.TextChoices):
    CREATED = 'created', _('Created but not queued')
    QUEUED = 'queued', _('Queued for processing')
    RUNNING = 'running', _('Processing')
    FAILED = 'failed', _('Failed')
    SUCCEEDED = 'success', _('Succeeded')
    SKIPPED = 'skipped', _('Skipped')


class TaskEventMixin(models.Model):
    """A mixin for models that must call a set of celery tasks.

    This mixin adds three class attributes:

    * ``task_funcs``, which should be the list of celery task functions that should be run
      on this model instance. Subclasses should set this attribute.
    * ``status``, a model field representing task execution status.
    * ``failure_reason``, a model field that can be set on this instance from within
      tasks for human-readable error logging.

    NOTE: you still need to register the pre/post save event.  on_commit events
    should be registered as post_save events, not pre_save.
    """

    class Meta:
        abstract = True

    failure_reason = models.TextField(null=True)
    status = models.CharField(max_length=20, default=Status.CREATED, choices=Status.choices)

    task_funcs: Iterable[Task] = []

    def _run_tasks(self) -> None:
        if not self.task_funcs:
            return

        self.status = Status.QUEUED
        self.save(
            update_fields=[
                'status',
            ]
        )
        for func in self.task_funcs:
            if getattr(settings, 'CELERY_TASK_ALWAYS_EAGER', None):
                # HACK: for some reason this is necessary
                func(self.pk)
            else:
                func.delay(self.pk)

    def _post_save_event_task(self, created: bool, *args, **kwargs) -> None:
        if not created and kwargs.get('update_fields'):
            return
        self._run_tasks()

    def _on_commit_event_task(self, *args, **kwargs) -> None:
        if kwargs.get('update_fields'):
            return
        self._run_tasks()


class DetailViewMixin:
    """Interface for spatial entry detail view redirect."""

    detail_view_name: str = None


class PermissionPathMixin:
    """Dummy model class for migrations."""

    ...



import logging
from typing import List

import djclick as click
from rgd.models import WhitelistedEmail

logger = logging.getLogger(__name__)


@click.command()
@click.option('-e', '--email', multiple=True)
def ingest_s3(
    email: List[str],
) -> None:
    for e in email:
        e = e.strip()
        logger.info(f'Whitelisting `{e}`')
        WhitelistedEmail.objects.get_or_create(email=e)



import json

from rest_framework import serializers

from . import models

MODIFIABLE_READ_ONLY_FIELDS = ['modified', 'created']
TASK_EVENT_READ_ONLY_FIELDS = ['status', 'failure_reason']
SPATIAL_ENTRY_EXCLUDE = ['footprint']


class RelatedField(serializers.PrimaryKeyRelatedField):
    """Handle GET/POST in a single field.

    Reference: https://stackoverflow.com/a/52246232
    """

    def __init__(self, **kwargs):
        self.serializer = kwargs.pop('serializer', None)
        if self.serializer is not None and not issubclass(self.serializer, serializers.Serializer):
            raise TypeError('"serializer" is not a valid serializer class')

        super().__init__(**kwargs)

    def use_pk_only_optimization(self):
        return False if self.serializer else True

    def to_representation(self, instance):
        if self.serializer:
            return self.serializer(instance, context=self.context).data
        return super().to_representation(instance)


class CollectionSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Collection
        fields = '__all__'

    def to_representation(self, instance):
        data = super().to_representation(instance)
        data['len'] = len(instance)
        return data


class CollectionPermissionSerializer(serializers.ModelSerializer):
    collection = RelatedField(
        queryset=models.Collection.objects.all(), serializer=CollectionSerializer
    )

    class Meta:
        model = models.CollectionPermission
        fields = '__all__'


class ChecksumFileFolderSerializer(serializers.Serializer):
    known_size = serializers.IntegerField()
    num_files = serializers.IntegerField()
    num_url_files = serializers.IntegerField()
    created = serializers.DateTimeField()
    modified = serializers.DateTimeField()


class ChecksumFileSerializer(serializers.ModelSerializer):
    """Serializer for ChecksumFiles.

    On POST, this can only handle URL files.
    """

    def to_representation(self, value):
        ret = super().to_representation(value)
        ret['download_url'] = value.get_url()
        ret['size'] = value.size
        return ret

    class Meta:
        model = models.ChecksumFile
        fields = '__all__'
        read_only_fields = (
            [
                'checksum',
                'last_validation',
            ]
            + MODIFIABLE_READ_ONLY_FIELDS
            + TASK_EVENT_READ_ONLY_FIELDS
        )


class ChecksumFilePathQuerySerializer(serializers.Serializer):
    path_prefix = serializers.CharField(required=False)


class ChecksumFilePathsSerializer(serializers.Serializer):
    folders = serializers.DictField(child=ChecksumFileFolderSerializer())
    files = serializers.DictField(child=ChecksumFileSerializer())


class ChecksumFileListQuerySerializer(serializers.Serializer):
    collection = serializers.IntegerField(required=False)
    url = serializers.CharField(required=False)


class SpatialEntrySerializer(serializers.ModelSerializer):
    outline = serializers.SerializerMethodField()
    subentry_name = serializers.SerializerMethodField()
    subentry_type = serializers.SerializerMethodField()

    class Meta:
        model = models.SpatialEntry
        exclude = SPATIAL_ENTRY_EXCLUDE

    def get_outline(self, obj):
        return json.loads(obj.outline.geojson)

    def get_subentry_type(self, obj):
        return type(obj).__name__

    def get_subentry_name(self, obj):
        return getattr(obj, 'name', None)


class SpatialEntryFootprintSerializer(SpatialEntrySerializer):
    footprint = serializers.SerializerMethodField()

    class Meta:
        model = models.SpatialEntry
        fields = '__all__'

    def get_footprint(self, obj):
        return json.loads(obj.footprint.geojson)


class SpatialAssetSerializer(SpatialEntrySerializer):
    file = RelatedField(
        queryset=models.ChecksumFile.objects.all(), serializer=ChecksumFileSerializer
    )

    class Meta:
        model = models.SpatialAsset
        exclude = SPATIAL_ENTRY_EXCLUDE



from django.contrib.gis.gdal import CoordTransform, SpatialReference
from django.contrib.gis.geos import GEOSGeometry
from rgd.models.constants import DB_SRID


def transform_geometry(geometry: GEOSGeometry, source_wkt: str):
    """Transform geometry into the database's spatial reference system."""
    source = SpatialReference(source_wkt)
    dest = SpatialReference(DB_SRID)
    transform = CoordTransform(source, dest)
    return geometry.transform(transform, clone=True)



from io import open as io_open
import os
from pathlib import Path

from setuptools import find_packages, setup

readme_file = Path(__file__).parent.parent / 'README.md'
if readme_file.exists():
    with readme_file.open() as f:
        long_description = f.read()
else:
    # When this is first installed in development Docker, README.md is not available
    long_description = ''

__version__ = None
filepath = os.path.dirname(__file__)
version_file = os.path.join(filepath, '..', 'version.py')
with io_open(version_file, mode='r') as fd:
    exec(fd.read())

setup(
    name='django-rgd',
    version=__version__,
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache 2.0',
    author='Kitware, Inc.',
    author_email='rgd@kitware.com',
    url='https://github.com/ResonantGeoData/ResonantGeoData',
    keywords='',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django :: 3.0',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python',
    ],
    python_requires='>=3.8',
    packages=find_packages(exclude=('client',)),
    include_package_data=True,
    dependency_links=[
        'https://girder.github.io/large_image_wheels',
    ],
    install_requires=[
        'boto3',
        'celery',
        'django>=3.2',  # See PR #264: due to this bug: https://code.djangoproject.com/ticket/31910
        'django-allauth',
        'django-click',
        'django-crum',
        'django-extensions',
        'django-filter',
        'django-girder-utils',
        'django-model-utils',
        'django-oauth-toolkit',
        'django-s3-file-field[minio]',
        'djangorestframework',
        'drf-yasg',
        'filelock',
        'flower',
        'GDAL',
        'pooch',
        'psutil',
        'psycopg2',
        'python-magic',
        'typing_extensions',
    ],
    extras_require={
        'fuse': [
            'numpy',  # requirement of `simple-httpfs`
            'simple-httpfs',
        ],
        'configuration': [
            'django-composed-configuration[dev]',
            'django-configurations[database,email]>=0.19.2',
            'django-cache-url',
            'redis',
        ],
    },
)



import copy

from django.contrib import admin
from django.contrib.gis.admin import OSMGeoAdmin

MODIFIABLE_FILTERS = (
    'modified',
    'created',
)

SPATIAL_ENTRY_FILTERS = (
    'acquisition_date',
    'instrumentation',
)

TASK_EVENT_FILTERS = ('status',)

TASK_EVENT_READONLY = (
    'failure_reason',
    'status',
)


class _FileGetNameMixin:
    def get_name(self, obj):
        return obj.file.name

    get_name.short_description = 'Name'
    get_name.admin_order_field = 'file__name'


def reprocess(modeladmin, request, queryset):
    """Trigger the save event task for each entry."""
    for entry in queryset.all():
        entry.save()


class GeoAdminInline(OSMGeoAdmin, admin.StackedInline):
    def __init__(self, parent_model, admin_site):
        self.admin_site = admin_site
        self.parent_model = parent_model
        self.opts = self.model._meta
        self.has_registered_model = admin_site.is_registered(self.model)
        overrides = copy.deepcopy(admin.options.FORMFIELD_FOR_DBFIELD_DEFAULTS)
        for k, v in self.formfield_overrides.items():
            overrides.setdefault(k, {}).update(v)
        self.formfield_overrides = overrides
        if self.verbose_name is None:
            self.verbose_name = self.model._meta.verbose_name
        if self.verbose_name_plural is None:
            self.verbose_name_plural = self.model._meta.verbose_name_plural



# Generated by Django 3.2.11 on 2022-02-01 17:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rgd', '0007_auto_20211215_0608'),
    ]

    operations = [
        migrations.AddField(
            model_name='collection',
            name='description',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.2.4 on 2021-06-24 22:52

from django.conf import settings
import django.contrib.gis.db.models.fields
from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import rgd.models.mixins
import rgd.utility
import s3_file_field.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Collection',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                ('name', models.CharField(max_length=127)),
            ],
            options={
                'default_related_name': 'collections',
            },
            bases=(models.Model, rgd.models.mixins.PermissionPathMixin),
        ),
        migrations.CreateModel(
            name='SpatialEntry',
            fields=[
                ('spatial_id', models.AutoField(primary_key=True, serialize=False)),
                ('acquisition_date', models.DateTimeField(blank=True, default=None, null=True)),
                ('footprint', django.contrib.gis.db.models.fields.GeometryField(srid=4326)),
                ('outline', django.contrib.gis.db.models.fields.GeometryField(srid=4326)),
                (
                    'instrumentation',
                    models.CharField(
                        blank=True,
                        help_text='The instrumentation used to acquire these data.',
                        max_length=100,
                        null=True,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name='WhitelistedEmail',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                ('email', models.EmailField(max_length=254)),
            ],
        ),
        migrations.CreateModel(
            name='CollectionPermission',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                (
                    'role',
                    models.SmallIntegerField(
                        choices=[(1, 'Reader'), (2, 'Owner')],
                        db_index=True,
                        default=1,
                        help_text='A "reader" can view assets in this collection. An "owner" can additionally add/remove other users, set their permissions, delete the collection, and add/remove other files.',
                    ),
                ),
                (
                    'collection',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='collection_permissions',
                        to='rgd.collection',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='collection_permissions',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'default_related_name': 'collection_permissions',
            },
        ),
        migrations.CreateModel(
            name='ChecksumFile',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name='ID'
                    ),
                ),
                ('failure_reason', models.TextField(null=True)),
                (
                    'status',
                    models.CharField(
                        choices=[
                            ('created', 'Created but not queued'),
                            ('queued', 'Queued for processing'),
                            ('running', 'Processing'),
                            ('failed', 'Failed'),
                            ('success', 'Succeeded'),
                        ],
                        default='created',
                        max_length=20,
                    ),
                ),
                (
                    'created',
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name='created'
                    ),
                ),
                (
                    'modified',
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name='modified'
                    ),
                ),
                ('name', models.CharField(blank=True, max_length=1000)),
                ('description', models.TextField(blank=True, null=True)),
                ('checksum', models.CharField(max_length=128)),
                ('validate_checksum', models.BooleanField(default=False)),
                ('last_validation', models.BooleanField(default=True)),
                ('type', models.IntegerField(choices=[(1, 'FileField'), (2, 'URL')], default=1)),
                (
                    'file',
                    s3_file_field.fields.S3FileField(
                        blank=True, null=True, upload_to=rgd.utility.uuid_prefix_filename
                    ),
                ),
                ('url', models.TextField(blank=True, null=True)),
                (
                    'collection',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name='checksumfiles',
                        related_query_name='checksumfiles',
                        to='rgd.collection',
                    ),
                ),
            ],
            bases=(models.Model, rgd.models.mixins.PermissionPathMixin),
        ),
        migrations.CreateModel(
            name='SpatialAsset',
            fields=[
                (
                    'spatialentry_ptr',
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to='rgd.spatialentry',
                    ),
                ),
                (
                    'created',
                    django_extensions.db.fields.CreationDateTimeField(
                        auto_now_add=True, verbose_name='created'
                    ),
                ),
                (
                    'modified',
                    django_extensions.db.fields.ModificationDateTimeField(
                        auto_now=True, verbose_name='modified'
                    ),
                ),
                ('files', models.ManyToManyField(to='rgd.ChecksumFile')),
            ],
            options={
                'get_latest_by': 'modified',
                'abstract': False,
            },
            bases=('rgd.spatialentry', models.Model, rgd.models.mixins.PermissionPathMixin),
        ),
        migrations.AddConstraint(
            model_name='collectionpermission',
            constraint=models.UniqueConstraint(fields=('collection', 'user'), name='unique_user'),
        ),
        migrations.AddConstraint(
            model_name='checksumfile',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        ('file__regex', '.+'),
                        ('type', 1),
                        models.Q(('url__in', ['', None]), ('url__isnull', True), _connector='OR'),
                    ),
                    models.Q(
                        ('type', 2),
                        models.Q(('url__isnull', False), ('url__regex', '.+')),
                        models.Q(('file__in', ['', None]), ('file__isnull', True), _connector='OR'),
                    ),
                    _connector='OR',
                ),
                name='rgd_checksumfile_file_source_value_matches_type',
            ),
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model
from django.core import signing
from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions


class UserSigner:
    """Signs/unsigns user object with an expiry."""

    signer_class = signing.TimestampSigner

    def sign(self, user):
        signer = self.signer_class()
        data = {'user_id': user.pk, 'username': user.get_username()}
        return signer.sign(signing.dumps(data))

    def unsign(self, signature, max_age=None):
        max_age = getattr(settings, 'RGD_SIGNED_URL_TTL', None)
        if max_age is None:
            max_age = 60 * 60 * 24
        cls = get_user_model()
        signer = self.signer_class()
        data = signing.loads(signer.unsign(signature, max_age))
        if not isinstance(data, dict):
            raise signing.BadSignature()
        try:
            return cls.objects.get(
                **{'pk': data.get('user_id'), cls.USERNAME_FIELD: data.get('username')}
            )
        except cls.DoesNotExist:
            raise signing.BadSignature()


class SignedURLAuthentication:
    """
    Extend the TokenAuthentication class to support signed authentication.

    This takes the form of "http://www.example.com/?signature=<key>".
    """

    def authenticate(self, request):
        # Check if 'signature' is in the request query params.
        # Give precedence to 'Authorization' header.
        param_name = getattr(settings, 'RGD_SIGNED_URL_QUERY_PARAM', 'signature')
        if param_name in request.query_params and 'HTTP_AUTHORIZATION' not in request.META:
            if request.method != 'GET':
                raise exceptions.AuthenticationFailed(
                    _('Signatures are only allowed for GET requests.')
                )
            signer = UserSigner()
            sig = request.query_params.get(param_name)
            if not sig:
                return
            try:
                user = signer.unsign(sig)
            except signing.SignatureExpired:
                raise exceptions.AuthenticationFailed(_('The signature has expired.'))
            except signing.BadSignature:
                raise exceptions.AuthenticationFailed(_('Invalid signature.'))
            if not user.is_active:
                raise exceptions.AuthenticationFailed(_('Signing user is inactive or deleted.'))
            return (user, None)



import io
from pathlib import Path
import tempfile

from django.db import IntegrityError
import pytest
from rgd.datastore import datastore, registry
from rgd.models import ChecksumFile, FileSourceType, utils
from rgd.models.collection import Collection

FILENAME = 'stars.png'


@pytest.fixture
def file_path():
    return datastore.fetch(FILENAME)


@pytest.mark.django_db(transaction=True)
def test_create_local_file(file_path):
    model = ChecksumFile()
    model.type = FileSourceType.FILE_FIELD
    with open(file_path, 'rb') as f:
        model.file.save(FILENAME, f)
    model.save()
    model.post_save_job()
    model.refresh_from_db()
    assert model.checksum == registry[FILENAME].split(':')[1]
    assert model.name == FILENAME


@pytest.mark.django_db(transaction=True)
def test_create_url():
    model = ChecksumFile()
    model.type = FileSourceType.URL
    model.url = datastore.get_url(FILENAME)
    model.save()
    model.post_save_job()
    model.refresh_from_db()
    assert model.checksum == registry[FILENAME].split(':')[1]
    assert model.name


@pytest.mark.django_db(transaction=True)
def test_create_url_local_file():
    path = Path('/opt/django-project/README.md')
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch()
    model = ChecksumFile()
    model.type = FileSourceType.URL
    model.url = f'file://{path}'
    model.save()
    model.post_save_job()
    model.refresh_from_db()
    assert model.checksum
    assert model.name


@pytest.mark.django_db(transaction=True)
def test_constraint_mismatch(file_path):
    with pytest.raises(IntegrityError):
        model = ChecksumFile()
        model.type = FileSourceType.FILE_FIELD
        model.url = datastore.get_url(FILENAME)
        model.save()
    with pytest.raises(IntegrityError):
        model = ChecksumFile()
        model.type = FileSourceType.URL
        with open(file_path, 'rb') as f:
            model.file.save(FILENAME, f)


@pytest.mark.django_db(transaction=True)
def test_constraint_url_null():
    with pytest.raises(IntegrityError):
        model = ChecksumFile()
        model.type = FileSourceType.URL
        model.save()


@pytest.mark.django_db(transaction=True)
def test_constraint_url_empty():
    with pytest.raises(IntegrityError):
        model = ChecksumFile()
        model.type = FileSourceType.URL
        model.url = ''  # empty string
        model.save()


@pytest.mark.django_db(transaction=True)
def test_constraint_file_with_empty_url(file_path):
    # Make sure the constraint passes when an empty string URL is given with
    #   the FileField choice. This happens when adding files in the admin interface
    model = ChecksumFile()
    model.type = FileSourceType.FILE_FIELD
    model.url = ''
    with open(file_path, 'rb') as f:
        model.file.save(FILENAME, f)
    assert not model.url
    assert model.file.name


@pytest.mark.django_db(transaction=True)
def test_get_or_create_file(file_path):
    with open(file_path, 'rb') as f:
        file, created = utils.get_or_create_checksumfile(file=f)
    assert created
    with open(file_path, 'rb') as f:
        file, created = utils.get_or_create_checksumfile(file=f)
    assert not created


@pytest.mark.django_db(transaction=True)
def test_get_or_create_file_permissions(file_path):
    collection = Collection.objects.create(name='Foo')
    with open(file_path, 'rb') as f:
        file, created = utils.get_or_create_checksumfile(collection=collection, file=f)
    assert created
    assert file.collection == collection
    with open(file_path, 'rb') as f:
        file, created = utils.get_or_create_checksumfile(collection=collection, file=f)
    assert not created
    with open(file_path, 'rb') as f:
        file, created = utils.get_or_create_checksumfile(file=f)
    # Because this passed collection is None, make sure a new file is created
    assert created
    assert file.collection is None


@pytest.mark.django_db(transaction=True)
def test_get_or_create_url():
    url = datastore.get_url(FILENAME)
    file, created = utils.get_or_create_checksumfile(url=url)
    assert created
    file, created = utils.get_or_create_checksumfile(url=url)
    assert not created


@pytest.mark.django_db(transaction=True)
def test_get_or_create_url_permissions():
    url = datastore.get_url(FILENAME)
    collection = Collection.objects.create(name='Foo')
    file, created = utils.get_or_create_checksumfile(collection=collection, url=url)
    assert created
    assert file.collection == collection
    file, created = utils.get_or_create_checksumfile(collection=collection, url=url)
    assert not created
    # Because this passed collection is None, make sure a new file is created
    file, created = utils.get_or_create_checksumfile(url=url)
    assert created
    assert file.collection is None


@pytest.mark.django_db(transaction=True)
def test_get_or_create_url_checksum():
    url = datastore.get_url(FILENAME)
    file, created = utils.get_or_create_checksumfile(url=url)
    assert created
    file, created = utils.get_or_create_checksumfile(url=url, precompute_url_checksum=True)
    assert not created


@pytest.mark.django_db(transaction=True)
def test_output_file_handler(settings, file_path):
    directory = tempfile.gettempdir()

    def output_file_handler(instance: ChecksumFile, file_handle: io.BufferedIOBase, name: str):
        # save to path on disk and give `file://` URL to ChecksumFile
        path = Path(directory, name)
        with open(path, 'wb') as f:
            f.write(file_handle.read())
        instance.type = FileSourceType.URL
        instance.url = f'file://{path}'
        instance.file = None

    settings.RGD_OUTPUT_FILE_HANDLER = output_file_handler

    file = ChecksumFile()
    with open(file_path, 'rb') as f:
        file.save_file_contents(f, FILENAME)
    file.save()
    assert file.url == f'file://{str(Path(directory, FILENAME))}'



from allauth.account.signals import user_signed_up
from django.conf import settings
from django.db import transaction
from django.db.models.signals import post_save
from django.dispatch import receiver
from rgd import models
from rgd.utility import skip_signal


@receiver(post_save, sender=models.ChecksumFile)
@skip_signal()
def _post_save_checksum_file(sender, instance, *args, **kwargs):
    transaction.on_commit(lambda: instance._post_save_event_task(*args, **kwargs))


@receiver(user_signed_up)
def set_new_user_inactive(sender, **kwargs):
    if getattr(settings, 'RGD_AUTO_APPROVE_SIGN_UP', None):
        # Use setting `RGD_AUTO_APPROVE_SIGN_UP` to automatically approve all users
        return
    user = kwargs.get('user')
    try:
        models.WhitelistedEmail.objects.get(email=user.email)
        user.is_active = True
    except models.WhitelistedEmail.DoesNotExist:
        user.is_active = False
    user.save(update_fields=['is_active'])



from contextlib import contextmanager
from importlib import import_module
import io
import logging
import os
from pathlib import Path
from urllib.error import URLError
from urllib.parse import urlparse

from crum import get_current_user
from django.conf import settings
from django.contrib.gis.db import models
from django_extensions.db.models import TimeStampedModel
from rgd.utility import (
    _link_url,
    clean_file_cache,
    compute_checksum_file_field,
    compute_checksum_url,
    compute_hash,
    download_field_file_to_local_path,
    download_url_file_to_local_path,
    get_cache_dir,
    get_file_lock,
    patch_internal_presign,
    precheck_fuse,
    safe_urlopen,
    url_file_to_fuse_path,
    uuid_prefix_filename,
)
from s3_file_field import S3FileField

from .. import tasks
from .collection import Collection
from .fileset import FileSet
from .mixins import TaskEventMixin

logger = logging.getLogger(__name__)


class FileSourceType(models.IntegerChoices):
    FILE_FIELD = 1, 'FileField'
    URL = 2, 'URL'


class ChecksumFile(TimeStampedModel, TaskEventMixin):
    """The main class for user-uploaded files.

    This has support for manually uploading files or specifying a URL to a file
    (for example in an existing S3 bucket). This broadly supports ``http<s>://``
    URLs to file resources as well as ``s3://`` as long as the node the app is
    running on is provisioned to access that S3 bucket.

    """

    name = models.CharField(max_length=1000, blank=True)
    description = models.TextField(null=True, blank=True)
    checksum = models.CharField(max_length=128)  # sha512
    validate_checksum = models.BooleanField(
        default=False
    )  # a flag to validate the checksum against the saved checksum
    last_validation = models.BooleanField(default=True)
    collection = models.ForeignKey(
        Collection,
        on_delete=models.SET_NULL,
        related_name='%(class)ss',
        related_query_name='%(class)ss',
        null=True,
        blank=True,
    )
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL, null=True, blank=True, default=None, on_delete=models.SET_NULL
    )
    file_set = models.ForeignKey(FileSet, null=True, blank=True, on_delete=models.SET_NULL)

    type = models.IntegerField(choices=FileSourceType.choices, default=FileSourceType.FILE_FIELD)
    file = S3FileField(null=True, blank=True, upload_to=uuid_prefix_filename)
    url = models.TextField(null=True, blank=True)

    task_funcs = (tasks.task_checksum_file_post_save,)

    class Meta:
        constraints = [
            models.CheckConstraint(
                name='%(app_label)s_%(class)s_file_source_value_matches_type',
                check=(
                    models.Q(
                        models.Q(type=FileSourceType.FILE_FIELD, file__regex=r'.+')
                        & models.Q(models.Q(url__in=['', None]) | models.Q(url__isnull=True))
                    )
                    | models.Q(
                        models.Q(type=FileSourceType.URL)
                        & models.Q(models.Q(url__isnull=False) & models.Q(url__regex=r'.+'))
                        & models.Q(models.Q(file__in=['', None]) | models.Q(file__isnull=True))
                    )
                ),
            ),
            models.UniqueConstraint(
                fields=['file_set', 'name'],
                name='unique_name',
            ),
            models.UniqueConstraint(
                fields=['collection', 'url'],
                name='unique_url_collection',
                condition=models.Q(type=FileSourceType.URL),  # Only apply if URL file
            ),
        ]

    @property
    def basename(self):
        return os.path.basename(self.name)

    @property
    def size(self):
        # Ensure safe check of self.file
        try:
            return self.file.size
        except ValueError:
            return None

    def get_checksum(self):
        """Compute a new checksum without saving it."""
        if self.type == FileSourceType.FILE_FIELD:
            return compute_checksum_file_field(self.file)
        elif self.type == FileSourceType.URL:
            parsed = urlparse(self.url)
            if parsed.scheme in ['https', 'http']:
                return compute_checksum_url(self.url)
            else:
                with self.yield_local_path() as path:
                    with open(path, 'rb') as f:
                        return compute_hash(f)
        else:
            raise NotImplementedError(f'Type ({self.type}) not supported.')

    def update_checksum(self):
        self.checksum = self.get_checksum()
        # Simple update save - not full save
        super(ChecksumFile, self).save(
            update_fields=[
                'checksum',
            ]
        )

    def validate(self):
        previous = self.checksum
        self.update_checksum()
        self.last_validation = self.checksum == previous
        # Simple update save - not full save
        super(ChecksumFile, self).save(
            update_fields=[
                'last_validation',
            ]
        )
        return self.last_validation

    def post_save_job(self):
        if not self.checksum or self.validate_checksum:
            if self.validate_checksum:
                self.validate()
            else:
                self.update_checksum()
            # Reset the user flags
            self.validate_checksum = False
            # Simple update save - not full save
            self.save(
                update_fields=[
                    'checksum',
                    'last_validation',
                    'validate_checksum',
                ]
            )

    def save(self, *args, **kwargs):
        if not self.name:
            if self.type == FileSourceType.FILE_FIELD and self.file.name:
                self.name = os.path.basename(self.file.name)
            elif self.type == FileSourceType.URL:
                parsed = urlparse(self.url)
                if parsed.scheme in ['https', 'http']:
                    try:
                        with safe_urlopen(self.url) as r:
                            self.name = r.info().get_filename()
                    except (AttributeError, ValueError, URLError):
                        pass
                if not self.name:
                    # Fallback
                    self.name = os.path.basename(urlparse(self.url).path)
        # Handle `created_by` User
        user = get_current_user()
        if user and not user.pk:
            user = None
        if not self.pk:
            self.created_by = user
        # Must save the model with the file before accessing it for the checksum
        super(ChecksumFile, self).save(*args, **kwargs)

    def download_to_local_path(self, directory: str = None):
        """Forcibly download this file to a directory on disk.

        Cleanup must be handled by caller.

        This will handle locking to prevent multiple processes/threads
        from trying to download the file at the same time -- only one thread
        or process will perform the download and the rest will yield its
        result.

        """
        if directory is None:
            dest_path = self.get_cache_path()
        else:
            dest_path = Path(directory, self.name)
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        # Thread/process safe locking for file access
        lock = get_file_lock(dest_path)

        with lock:  # TODO: handle timeouts in condition
            if dest_path.exists() and dest_path.stat().st_size > 0:
                # File already exists (is cached)
                logger.debug(f'Found cached file ({self.pk}) at: {dest_path}')
                # Touch the file so that it moves to the top of the priority list
                # when cleaning.
                if not os.path.islink(str(dest_path)):
                    dest_path.touch()
                return dest_path
            else:
                logger.debug(f'Downloading file ({self.pk}) to: {dest_path}')
                # If downloading to the cache, clean to achieve available free space
                if get_cache_dir() in dest_path.parents:
                    clean_file_cache()
                # TODO: handle if these fail (e.g. bad S3 credentials)
                if self.type == FileSourceType.FILE_FIELD:
                    return download_field_file_to_local_path(self.file, dest_path)
                elif self.type == FileSourceType.URL:
                    return download_url_file_to_local_path(self.url, dest_path)

    def get_cache_path(self, root: bool = False):
        """Generate a predetermined path in the cache directory.

        This will use the associated FileSet's cache path if this resource
        has a file_set, otherwise it will place in the top of the cache
        directory.

        """
        if self.file_set is None:
            # If no file_set, place in the main cache directory
            directory = get_cache_dir() / f'f-{self.pk}'
            directory.mkdir(parents=True, exist_ok=True)
        else:
            directory = self.file_set.get_cache_path()
        if root:
            return directory
        return directory / f'{self.name}'

    @contextmanager
    def yield_local_path(self, try_fuse: bool = True, yield_file_set: bool = False):
        """Create a local path for this file and all other files in its file_set.

        This will first attempt to use httpfs to FUSE mount the file's URL if
        and only if the file does not belong to a FileSet. FUSE with multiple
        files in a FileSet is not yet supported.

        If FUSE is unavailable, this will fallback to downloading the entire
        file (and the other files in this item's FileSet) to local storage.

        Parameters
        ----------
        try_fuse : bool
            Try to use the FUSE interface. If false, use VSI or download to
            local storage.

        yield_file_set : bool
            Yield all of the files in this file's file_set if available.

        """
        # TODO: fix FUSE to handle adjacent files
        if (
            self.file_set is None
            and try_fuse
            and self.type == FileSourceType.URL
            and precheck_fuse(self.get_url())
        ):
            yield url_file_to_fuse_path(self.get_url(internal=True))
            return
        # Fallback to loading entire file locally - this uses `get_temp_path`
        logger.debug('`yield_local_path` falling back to downloading entire file to local storage.')
        root = self.get_cache_path(root=True)
        path = self.get_cache_path()
        if yield_file_set and self.file_set:
            # NOTE: This is messy and should be improved but it ensures the directory remains locked
            with self.file_set.yield_all_to_local_path() as _:
                yield path
            return
        # Not in file_set. Download to cache dir
        from .utils import yield_checksumfiles

        with yield_checksumfiles([self], root):
            yield path

    def get_url(self, internal: bool = False):
        """Get the URL of the stored resource.

        Parameters
        ----------
        internal : bool
            In most cases this URL will be accessible from anywhere. In some
            cases, this URL will only be accessible from within the container.
            This flag is for use with internal processes to make sure the host
            is correctly set to ``minio`` when needed. See
            ``patch_internal_presign`` for more details.

        """
        if self.type == FileSourceType.FILE_FIELD:
            if internal:
                with patch_internal_presign(self.file):
                    return self.file.url
            else:
                return self.file.url
        elif self.type == FileSourceType.URL:
            return self.url

    def data_link(self):
        return _link_url(self, 'get_url')

    data_link.allow_tags = True

    def save_file_contents(self, file_handle: io.BufferedIOBase, name: str):
        if hasattr(settings, 'RGD_OUTPUT_FILE_HANDLER'):
            if isinstance(settings.RGD_OUTPUT_FILE_HANDLER, str):
                parts = settings.RGD_OUTPUT_FILE_HANDLER.rsplit('.', 1)
                func = getattr(import_module(parts[0]), parts[1])
            elif callable(settings.RGD_OUTPUT_FILE_HANDLER):
                func = settings.RGD_OUTPUT_FILE_HANDLER
            else:
                raise ValueError(
                    f'`output_file_handler` not understood: {settings.RGD_OUTPUT_FILE_HANDLER}'
                )
            func(self, file_handle, name)
        else:
            if not self.name:
                self.name = name
            self.type = FileSourceType.FILE_FIELD
            self.file.save(name, file_handle)
        if self.pk:
            self.save(update_fields=['type', 'file', 'url', 'name'])
        else:
            self.save()



from .collection import Collection, CollectionPermission  # noqa
from .common import SpatialAsset, SpatialEntry, WhitelistedEmail  # noqa
from .constants import *  # noqa
from .file import ChecksumFile, FileSourceType  # noqa
from .fileset import FileSet  # noqa
from .mixins import *  # noqa
from .transform import transform_geometry  # noqa
